# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

import platform
from ctypes import CDLL, c_char_p, c_int, c_void_p

from ..utils import find_data_file, u2b

_libfile = {
    "Linux": "libepScriptLib.so",  # Linux
    "Windows": "libepScriptLib.dll",  # Windows
    "Darwin": "libepScriptLib.dylib",  # Mac
}[platform.system()]


libeps = CDLL(find_data_file(_libfile, __file__))
libeps.compileString.argtypes = [c_char_p, c_char_p]
libeps.compileString.restype = c_void_p
libeps.freeCompiledResult.argtypes = [c_void_p]
libeps.setDebugMode.argtypes = [c_int]
libeps.getErrorCount.argtypes = []
libeps.getErrorCount.restype = c_int
libeps.registerPlibConstants.argtypes = [c_char_p]
libeps.registerPyKeywords.argtypes = [c_char_p]
libeps.registerPyBuiltins.argtypes = [c_char_p]


def _set_eps_globals(global_list):
    global_list_c = b"\0".join(u2b(g) for g in global_list) + b"\0"
    libeps.registerPlibConstants(global_list_c)


def _set_py_keywords(keyword_list):
    keyword_list_c = b"\0".join(u2b(g) for g in keyword_list) + b"\0"
    libeps.registerPyKeywords(keyword_list_c)


def _set_py_builtins(builtin_list):
    builtin_list_c = b"\0".join(u2b(g) for g in builtin_list) + b"\0"
    libeps.registerPyBuiltins(builtin_list_c)


def epsCompile(filename, b_code):  # noqa: N802
    filename = u2b(filename)
    output = libeps.compileString(filename, b_code)
    if not output or libeps.getErrorCount():
        return None
    output_str = c_char_p(output).value
    libeps.freeCompiledResult(output)
    return output_str


def EPS_SetDebug(b):  # noqa: N802
    libeps.setDebugMode(b)



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from ..localize import _
from .eperror import EPError

g_encoding: str = "cp949"


def u2b(s: str | bytes) -> bytes:
    if isinstance(s, str):
        try:
            return s.encode(g_encoding)
        except UnicodeEncodeError:
            return s.encode("utf-8")
    elif isinstance(s, bytes):
        return s
    else:
        raise EPError(_("Invalid type {}").format(type(s)))


def b2u(b: str | bytes) -> str:
    if isinstance(b, bytes):
        return b.decode(g_encoding)
    elif isinstance(b, str):
        return b
    else:
        raise EPError(_("Invalid type {}").format(type(b)))


def u2utf8(s: str | bytes) -> bytes:
    if isinstance(s, str):
        return s.encode("utf-8")
    elif isinstance(s, bytes):
        return s
    else:
        raise EPError(_("Invalid type {}").format(type(s)))


def b2utf8(b: str | bytes) -> str:
    if isinstance(b, bytes):
        return b.decode("utf-8")
    elif isinstance(b, str):
        return b
    else:
        raise EPError(_("Invalid type {}").format(type(b)))



from typing import Literal, TypeAlias

DefSpriteDict: dict[str, int] = {
    "2/38 Ash": 0,
    "2/39 Ash": 1,
    "2/41 Ash": 2,
    "2/40 Ash": 3,
    "2/42 Ash": 4,
    "2/43 Ash": 5,
    "2/44 Ash": 6,
    "2/1 Ash": 7,
    "2/4 Ash": 8,
    "2/5 Ash": 9,
    "2/30 Ash": 10,
    "2/28 Ash": 11,
    "2/29 Ash": 12,
    "4/1 Ash": 13,
    "4/2 Ash": 14,
    "4/3 Ash": 15,
    "4/56 Jungle": 16,
    "4/57 Jungle": 17,
    "4/58 Jungle": 18,
    "4/59 Jungle": 19,
    "9/5 Jungle": 20,
    "9/6 Jungle": 21,
    "9/7 Jungle": 22,
    "4/51 Jungle": 23,
    "4/52 Jungle": 24,
    "4/54 Jungle": 25,
    "4/53 Jungle": 26,
    "9/1 Jungle": 27,
    "9/2 Jungle": 28,
    "9/3 Jungle": 29,
    "9/4 Jungle": 30,
    "4/12 Jungle": 31,
    "4/13 Jungle": 32,
    "4/1 Jungle": 33,
    "4/3 Jungle": 34,
    "4/2 Jungle": 35,
    "4/5 Jungle": 36,
    "4/4 Jungle": 37,
    "4/9 Jungle": 38,
    "4/10 Jungle": 39,
    "5/5 Jungle": 40,
    "5/7 Jungle": 41,
    "5/6 Jungle": 42,
    "5/9 Jungle": 43,
    "5/8 Jungle": 44,
    "4/6 Jungle": 45,
    "4/7 Jungle": 46,
    "4/17 Jungle": 47,
    "13/4 Jungle": 48,
    "11/5 Jungle": 49,
    "11/6 Jungle": 50,
    "11/7 Jungle": 51,
    "11/8 Jungle": 52,
    "11/10 Jungle": 53,
    "11/11 Jungle": 54,
    "11/12 Jungle": 55,
    "7/4 Platform": 56,
    "7/5 Platform": 57,
    "7/6 Platform": 58,
    "7/1 Platform": 59,
    "7/2 Platform": 60,
    "7/3 Platform": 61,
    "7/9 Platform": 62,
    "7/10 Platform": 63,
    "7/8 Platform": 64,
    "7/7 Platform": 65,
    "7/26 Platform": 66,
    "7/24 Platform": 67,
    "7/28 Platform": 68,
    "7/27 Platform": 69,
    "7/25 Platform": 70,
    "7/29 Platform": 71,
    "7/30 Platform": 72,
    "7/31 Platform": 73,
    "12/1 Platform": 74,
    "9/27 Platform": 75,
    "5/54 Badlands": 76,
    "5/55 Badlands": 77,
    "5/56 Badlands": 78,
    "5/57 Badlands": 79,
    "6/16 Badlands": 80,
    "6/17 Badlands": 81,
    "6/20 Badlands": 82,
    "6/21 Badlands": 83,
    "5/10 Badlands": 84,
    "5/50 Badlands": 85,
    "5/52 Badlands": 86,
    "5/53 Badlands": 87,
    "5/51 Badlands": 88,
    "6/3 Badlands": 89,
    "11/3 Badlands": 90,
    "11/8 Badlands": 91,
    "11/6 Badlands": 92,
    "11/7 Badlands": 93,
    "11/9 Badlands": 94,
    "11/10 Badlands": 95,
    "11/11 Badlands": 96,
    "11/12 Badlands": 97,
    "11/13 Badlands": 98,
    "11/14 Badlands": 99,
    "1/13 Badlands": 100,
    "1/9 Badlands": 101,
    "1/11 Badlands": 102,
    "1/14 Badlands": 103,
    "1/10 Badlands": 104,
    "1/12 Badlands": 105,
    "1/15 Badlands": 106,
    "1/7 Badlands": 107,
    "1/5 Badlands": 108,
    "1/16 Badlands": 109,
    "1/8 Badlands": 110,
    "1/6 Badlands1": 111,
    "1/6 Badlands2": 112,
    "1/6 Badlands3": 113,
    "1/6 Badlands4": 114,
    "1/6 Badlands5": 115,
    "1/6 Badlands6": 116,
    "1/6 Badlands7": 117,
    "1/6 Badlands8": 118,
    "4/15 Installation1": 119,
    "4/15 Installation2": 120,
    "3/9 Installation": 121,
    "3/10 Installation": 122,
    "3/11 Installation": 123,
    "3/12 Installation": 124,
    "1/6 Badlands9": 125,
    "1/6 Badlands10": 126,
    "3/1 Installation": 127,
    "3/2 Installation": 128,
    "1/6 Badlands11": 129,
    "Scourge": 130,
    "Scourge Death": 131,
    "Scourge Explosion": 132,
    "Broodling": 133,
    "Broodling Remnants": 134,
    "Infested Terran": 135,
    "Infested Terran Explosion": 136,
    "Guardian Cocoon": 137,
    "Defiler": 138,
    "Defiler Remnants": 139,
    "Drone": 140,
    "Drone Remnants": 141,
    "Egg": 142,
    "Egg Remnants": 143,
    "Guardian": 144,
    "Guardian Death": 145,
    "Hydralisk": 146,
    "Hydralisk Remnants": 147,
    "Infested Kerrigan": 148,
    "Larva": 149,
    "Larva Remnants": 150,
    "Mutalisk": 151,
    "Mutalisk Death": 152,
    "Overlord": 153,
    "Overlord Death": 154,
    "Queen": 155,
    "Queen Death": 156,
    "Ultralisk": 157,
    "Ultralisk Remnants": 158,
    "Zergling": 159,
    "Zergling Remnants": 160,
    "Cerebrate": 161,
    "Infested Command Center": 162,
    "Spawning Pool": 163,
    "Mature Chrysalis": 164,
    "Evolution Chamber": 165,
    "Creep Colony": 166,
    "Hatchery": 167,
    "Hive": 168,
    "Lair": 169,
    "Sunken Colony": 170,
    "Greater Spire": 171,
    "Defiler Mound": 172,
    "Queen's Nest": 173,
    "Nydus Canal": 174,
    "Overmind With Shell": 175,
    "Overmind Without Shell": 176,
    "Ultralisk Cavern": 177,
    "Extractor": 178,
    "Hydralisk Den": 179,
    "Spire": 180,
    "Spore Colony": 181,
    "Zerg Building Spawn (Small)": 182,
    "Zerg Building Spawn (Medium)": 183,
    "Zerg Building Spawn (Large)": 184,
    "Zerg Building Explosion": 185,
    "Zerg Building Rubble (Small)": 186,
    "Zerg Building Rubble (Large)": 187,
    "Arbiter": 188,
    "Archon Energy": 189,
    "Carrier": 190,
    "Dragoon": 191,
    "Dragoon Remnants": 192,
    "Interceptor": 193,
    "Probe": 194,
    "Scout": 195,
    "Shuttle": 196,
    "High Templar": 197,
    "Dark Templar (Hero)": 198,
    "Reaver": 199,
    "Scarab": 200,
    "Zealot": 201,
    "Observer": 202,
    "Templar Archives": 203,
    "Assimilator": 204,
    "Observatory": 205,
    "Citadel of Adun": 206,
    "Forge": 207,
    "Gateway": 208,
    "Cybernetics Core": 209,
    "Khaydarin Crystal Formation": 210,
    "Nexus": 211,
    "Photon Cannon": 212,
    "Arbiter Tribunal": 213,
    "Pylon": 214,
    "Robotics Facility": 215,
    "Shield Battery": 216,
    "Stargate": 217,
    "Stasis Cell/Prison": 218,
    "Robotics Support Bay": 219,
    "Protoss Temple": 220,
    "Fleet Beacon": 221,
    "Explosion (Large)": 222,
    "Protoss Building Rubble (Small)": 223,
    "Protoss Building Rubble (Large)": 224,
    "Battlecruiser": 225,
    "Civilian": 226,
    "Dropship": 227,
    "Firebat": 228,
    "Ghost": 229,
    "Ghost Remnants": 230,
    "Nuke Target Dot": 231,
    "Goliath Base": 232,
    "Goliath Turret": 233,
    "Sarah Kerrigan": 234,
    "Marine": 235,
    "Marine Remnants": 236,
    "Scanner Sweep": 237,
    "Wraith": 238,
    "SCV": 239,
    "Siege Tank (Tank) Base": 240,
    "Siege Tank (Tank) Turret": 241,
    "Siege Tank (Siege) Base": 242,
    "Siege Tank (Siege) Turret": 243,
    "Vulture": 244,
    "Spider Mine": 245,
    "Science Vessel (Base)": 246,
    "Science Vessel (Turret)": 247,
    "Terran Academy": 248,
    "Barracks": 249,
    "Armory": 250,
    "Comsat Station": 251,
    "Command Center": 252,
    "Supply Depot": 253,
    "Control Tower": 254,
    "Factory": 255,
    "Covert Ops": 256,
    "Ion Cannon": 257,
    "Machine Shop": 258,
    "Missile Turret (Base)": 259,
    "Crashed Batlecruiser": 260,
    "Physics Lab": 261,
    "Bunker": 262,
    "Refinery": 263,
    "Science Facility": 264,
    "Nuclear Silo": 265,
    "Nuclear Missile": 266,
    "Nuke Hit": 267,
    "Starport": 268,
    "Engineering Bay": 269,
    "Terran Construction (Large)": 270,
    "Terran Construction (Small)": 271,
    "Building Explosion (Large)": 272,
    "Terran Building Rubble (Small)": 273,
    "Terran Building Rubble (Large)": 274,
    "Vespene Geyser": 275,
    "Ragnasaur (Ash)": 276,
    "Rhynadon (Badlands)": 277,
    "Bengalaas (Jungle)": 278,
    "Mineral Field Type1": 279,
    "Mineral Field Type2": 280,
    "Mineral Field Type3": 281,
    "Independent Starport (Unused)": 282,
    "Zerg Beacon": 283,
    "Terran Beacon": 284,
    "Protoss Beacon": 285,
    "Dark Swarm": 286,
    "Flag": 287,
    "Young Chrysalis": 288,
    "Psi Emitter": 289,
    "Data Disc": 290,
    "Khaydarin Crystal": 291,
    "Mineral Chunk Type1": 292,
    "Mineral Chunk Type2": 293,
    "Protoss Gas Orb Type1": 294,
    "Protoss Gas Orb Type2": 295,
    "Zerg Gas Sac Type1": 296,
    "Zerg Gas Sac Type2": 297,
    "Terran Gas Tank Type1": 298,
    "Terran Gas Tank Type2": 299,
    "White Circle (Invisible)": 300,
    "Start Location": 301,
    "Map Revealer": 302,
    "Floor Gun Trap": 303,
    "Wall Missile Trap": 304,
    "Wall Missile Trap2": 305,
    "Wall Flame Trap": 306,
    "Wall Flame Trap2": 307,
    "Floor Missile Trap": 308,
    "Longbolt/Gemini Missiles Trail": 309,
    "Grenade Shot Smoke": 310,
    "Vespene Geyser Smoke1": 311,
    "Vespene Geyser Smoke2": 312,
    "Vespene Geyser Smoke3": 313,
    "Vespene Geyser Smoke4": 314,
    "Vespene Geyser Smoke5": 315,
    "Small Explosion (Unused)": 316,
    "Double Explosion": 317,
    "Cursor Marker": 318,
    "Egg Spawn": 319,
    "High Templar Glow": 320,
    "Psi Field (Right Upper)": 321,
    "Burrowing Dust": 322,
    "Building Landing Dust Type1": 323,
    "Building Landing Dust Type2": 324,
    "Building Landing Dust Type3": 325,
    "Building Landing Dust Type4": 326,
    "Building Landing Dust Type5": 327,
    "Building Lifting Dust Type1": 328,
    "Building Lifting Dust Type2": 329,
    "Building Lifting Dust Type3": 330,
    "Building Lifting Dust Type4": 331,
    "Needle Spines": 332,
    "Dual Photon Blasters Hit": 333,
    "Particle Beam Hit": 334,
    "Anti-Matter Missile": 335,
    "Pulse Cannon": 336,
    "Phase Disruptor": 337,
    "STA/STS Photon Cannon Overlay": 338,
    "Psionic Storm": 339,
    "Fusion Cutter Hit": 340,
    "Gauss Rifle Hit": 341,
    "Gemini Missiles": 342,
    "Fragmentation Grenade": 343,
    "Magna Pulse (Unused)": 344,
    "Lockdown/LongBolt/Hellfire Missile": 345,
    "C-10 Canister Rifle Hit": 346,
    "ATS/ATA Laser Battery": 347,
    "Burst Lasers": 348,
    "Arclite Shock Cannon Hit": 349,
    "Yamato Gun": 350,
    "Yamato Gun Trail": 351,
    "EMP Shockwave Missile": 352,
    "Needle Spine Hit": 353,
    "Plasma Drip Hit (Unused)": 354,
    "Sunken Colony Tentacle": 355,
    "Venom (Unused Zerg Weapon)": 356,
    "Acid Spore": 357,
    "Glave Wurm": 358,
    "Seeker Spores": 359,
    "Queen Spell Holder": 360,
    "Stasis Field Hit": 361,
    "Plague Cloud": 362,
    "Consume": 363,
    "Ensnare": 364,
    "Glave Wurm/Seeker Spores Hit": 365,
    "Psionic Shockwave Hit": 366,
    "Glave Wurm Trail": 367,
    "Seeker Spores Overlay": 368,
    "Phase Disruptor (Unused)": 369,
    "White Circle": 370,
    "Acid Spray (Unused)": 371,
    "Plasma Drip (Unused)": 372,
    "Scarab/Anti-Matter Missile Overlay": 373,
    "Hallucination Death1": 374,
    "Hallucination Death2": 375,
    "Hallucination Death3": 376,
    "Bunker Overlay": 377,
    "FlameThrower": 378,
    "Recall Field": 379,
    "Scanner Sweep Hit": 380,
    "Left Upper Level Door": 381,
    "Right Upper Level Door": 382,
    "Substructure Left Door": 383,
    "Substructure Right Door": 384,
    "Substructure Opening Hole": 385,
    "7/13 Twilight": 386,
    "7/14 Twilight": 387,
    "7/16 Twilight": 388,
    "7/15 Twilight": 389,
    "7/19 Twilight": 390,
    "7/20 Twilight": 391,
    "7/21 Twilight": 392,
    "Unknown Twilight": 393,
    "7/17 Twilight": 394,
    "6/1 Twilight": 395,
    "6/2 Twilight": 396,
    "6/3 Twilight": 397,
    "6/4 Twilight": 398,
    "6/5 Twilight": 399,
    "8/3 Twilight": 400,
    "8/4 Twilight": 401,
    "9/29 Ice": 402,
    "9/28 Ice": 403,
    "12/38 Ice": 404,
    "12/37 Ice": 405,
    "12/33 Ice": 406,
    "9/21 Ice": 407,
    "9/15 Ice": 408,
    "9/16 Ice": 409,
    "Unknown410": 410,
    "Unknown411": 411,
    "12/9 Ice1": 412,
    "12/10 Ice": 413,
    "9/24 Ice": 414,
    "9/23 Ice": 415,
    "Unknown416": 416,
    "12/7 Ice": 417,
    "12/8 Ice": 418,
    "12/9 Ice2": 419,
    "12/11 Ice": 420,
    "12/40 Ice": 421,
    "12/41 Ice": 422,
    "12/42 Ice": 423,
    "12/5 Ice": 424,
    "12/6 Ice": 425,
    "12/36 Ice": 426,
    "12/32 Ice": 427,
    "12/35 Ice": 428,
    "12/34 Ice": 429,
    "12/24 Ice": 430,
    "12/25 Ice": 431,
    "9/22 Ice": 432,
    "12/31 Ice": 433,
    "12/20 Ice": 434,
    "12/26 Ice": 435,
    "12/27 Ice": 436,
    "12/30 Ice": 437,
    "12/39 Ice": 438,
    "Unknown439": 439,
    "4/1 Ice": 440,
    "6/1 Ice": 441,
    "5/6 Ice": 442,
    "5/7 Ice": 443,
    "5/8 Ice": 444,
    "5/9 Ice": 445,
    "10/10 Desert1": 446,
    "10/12 Desert1": 447,
    "10/8 Desert": 448,
    "10/9 Desert1": 449,
    "6/10 Desert": 450,
    "6/13 Desert": 451,
    "Unknown Desert": 452,
    "10/12 Desert2": 453,
    "10/9 Desert2": 454,
    "10/10 Desert2": 455,
    "10/11 Desert": 456,
    "10/14 Desert": 457,
    "10/41 Desert": 458,
    "10/39 Desert": 459,
    "10/9 Desert": 460,
    "10/6 Desert": 461,
    "10/7 Desert": 462,
    "4/6 Desert": 463,
    "4/11 Desert": 464,
    "4/10 Desert": 465,
    "4/9 Desert": 466,
    "4/7 Desert": 467,
    "4/12 Desert": 468,
    "4/8 Desert": 469,
    "4/13 Desert": 470,
    "6/14 Desert": 471,
    "4/17 Desert": 472,
    "6/20 Desert": 473,
    "4/15 Desert1": 474,
    "4/15 Desert2": 475,
    "10/23 Desert": 476,
    "8/23 Desert": 477,
    "10/5 Desert": 478,
    "12/1 Desert Overlay": 479,
    "11/3 Desert": 480,
    "Lurker Egg": 481,
    "Devourer": 482,
    "Devourer Death": 483,
    "Lurker Remnants": 484,
    "Lurker": 485,
    "Dark Archon Energy": 486,
    "Corsair": 487,
    "Dark Templar (Unit)": 488,
    "Medic": 489,
    "Medic Remnants": 490,
    "Valkyrie": 491,
    "Scantid (Desert)": 492,
    "Kakaru (Twilight)": 493,
    "Ursadon (Ice)": 494,
    "Overmind Cocoon": 495,
    "Power Generator": 496,
    "Xel'Naga Temple": 497,
    "Psi Disrupter": 498,
    "Warp Gate": 499,
    "Feedback Hit (Small)": 500,
    "Feedback Hit (Medium)": 501,
    "Feedback Hit (Large)": 502,
    "Disruption Web": 503,
    "White Circle2": 504,
    "Halo Rockets Trail": 505,
    "Neutron Flare": 506,
    "Neutron Flare Overlay (Unused)": 507,
    "Optical Flare Grenade": 508,
    "Halo Rockets": 509,
    "Subterranean Spines Hit": 510,
    "Subterranean Spines": 511,
    "Corrosive Acid Shot": 512,
    "Corrosive Acid Hit": 513,
    "Maelstrom Hit": 514,
    "Uraj": 515,
    "Khalis": 516,
}

DefaultSprite: TypeAlias = Literal[
    "2/38 Ash",
    "2/39 Ash",
    "2/41 Ash",
    "2/40 Ash",
    "2/42 Ash",
    "2/43 Ash",
    "2/44 Ash",
    "2/1 Ash",
    "2/4 Ash",
    "2/5 Ash",
    "2/30 Ash",
    "2/28 Ash",
    "2/29 Ash",
    "4/1 Ash",
    "4/2 Ash",
    "4/3 Ash",
    "4/56 Jungle",
    "4/57 Jungle",
    "4/58 Jungle",
    "4/59 Jungle",
    "9/5 Jungle",
    "9/6 Jungle",
    "9/7 Jungle",
    "4/51 Jungle",
    "4/52 Jungle",
    "4/54 Jungle",
    "4/53 Jungle",
    "9/1 Jungle",
    "9/2 Jungle",
    "9/3 Jungle",
    "9/4 Jungle",
    "4/12 Jungle",
    "4/13 Jungle",
    "4/1 Jungle",
    "4/3 Jungle",
    "4/2 Jungle",
    "4/5 Jungle",
    "4/4 Jungle",
    "4/9 Jungle",
    "4/10 Jungle",
    "5/5 Jungle",
    "5/7 Jungle",
    "5/6 Jungle",
    "5/9 Jungle",
    "5/8 Jungle",
    "4/6 Jungle",
    "4/7 Jungle",
    "4/17 Jungle",
    "13/4 Jungle",
    "11/5 Jungle",
    "11/6 Jungle",
    "11/7 Jungle",
    "11/8 Jungle",
    "11/10 Jungle",
    "11/11 Jungle",
    "11/12 Jungle",
    "7/4 Platform",
    "7/5 Platform",
    "7/6 Platform",
    "7/1 Platform",
    "7/2 Platform",
    "7/3 Platform",
    "7/9 Platform",
    "7/10 Platform",
    "7/8 Platform",
    "7/7 Platform",
    "7/26 Platform",
    "7/24 Platform",
    "7/28 Platform",
    "7/27 Platform",
    "7/25 Platform",
    "7/29 Platform",
    "7/30 Platform",
    "7/31 Platform",
    "12/1 Platform",
    "9/27 Platform",
    "5/54 Badlands",
    "5/55 Badlands",
    "5/56 Badlands",
    "5/57 Badlands",
    "6/16 Badlands",
    "6/17 Badlands",
    "6/20 Badlands",
    "6/21 Badlands",
    "5/10 Badlands",
    "5/50 Badlands",
    "5/52 Badlands",
    "5/53 Badlands",
    "5/51 Badlands",
    "6/3 Badlands",
    "11/3 Badlands",
    "11/8 Badlands",
    "11/6 Badlands",
    "11/7 Badlands",
    "11/9 Badlands",
    "11/10 Badlands",
    "11/11 Badlands",
    "11/12 Badlands",
    "11/13 Badlands",
    "11/14 Badlands",
    "1/13 Badlands",
    "1/9 Badlands",
    "1/11 Badlands",
    "1/14 Badlands",
    "1/10 Badlands",
    "1/12 Badlands",
    "1/15 Badlands",
    "1/7 Badlands",
    "1/5 Badlands",
    "1/16 Badlands",
    "1/8 Badlands",
    "1/6 Badlands1",
    "1/6 Badlands2",
    "1/6 Badlands3",
    "1/6 Badlands4",
    "1/6 Badlands5",
    "1/6 Badlands6",
    "1/6 Badlands7",
    "1/6 Badlands8",
    "4/15 Installation1",
    "4/15 Installation2",
    "3/9 Installation",
    "3/10 Installation",
    "3/11 Installation",
    "3/12 Installation",
    "1/6 Badlands9",
    "1/6 Badlands10",
    "3/1 Installation",
    "3/2 Installation",
    "1/6 Badlands11",
    "Scourge",
    "Scourge Death",
    "Scourge Explosion",
    "Broodling",
    "Broodling Remnants",
    "Infested Terran",
    "Infested Terran Explosion",
    "Guardian Cocoon",
    "Defiler",
    "Defiler Remnants",
    "Drone",
    "Drone Remnants",
    "Egg",
    "Egg Remnants",
    "Guardian",
    "Guardian Death",
    "Hydralisk",
    "Hydralisk Remnants",
    "Infested Kerrigan",
    "Larva",
    "Larva Remnants",
    "Mutalisk",
    "Mutalisk Death",
    "Overlord",
    "Overlord Death",
    "Queen",
    "Queen Death",
    "Ultralisk",
    "Ultralisk Remnants",
    "Zergling",
    "Zergling Remnants",
    "Cerebrate",
    "Infested Command Center",
    "Spawning Pool",
    "Mature Chrysalis",
    "Evolution Chamber",
    "Creep Colony",
    "Hatchery",
    "Hive",
    "Lair",
    "Sunken Colony",
    "Greater Spire",
    "Defiler Mound",
    "Queen's Nest",
    "Nydus Canal",
    "Overmind With Shell",
    "Overmind Without Shell",
    "Ultralisk Cavern",
    "Extractor",
    "Hydralisk Den",
    "Spire",
    "Spore Colony",
    "Zerg Building Spawn (Small)",
    "Zerg Building Spawn (Medium)",
    "Zerg Building Spawn (Large)",
    "Zerg Building Explosion",
    "Zerg Building Rubble (Small)",
    "Zerg Building Rubble (Large)",
    "Arbiter",
    "Archon Energy",
    "Carrier",
    "Dragoon",
    "Dragoon Remnants",
    "Interceptor",
    "Probe",
    "Scout",
    "Shuttle",
    "High Templar",
    "Dark Templar (Hero)",
    "Reaver",
    "Scarab",
    "Zealot",
    "Observer",
    "Templar Archives",
    "Assimilator",
    "Observatory",
    "Citadel of Adun",
    "Forge",
    "Gateway",
    "Cybernetics Core",
    "Khaydarin Crystal Formation",
    "Nexus",
    "Photon Cannon",
    "Arbiter Tribunal",
    "Pylon",
    "Robotics Facility",
    "Shield Battery",
    "Stargate",
    "Stasis Cell/Prison",
    "Robotics Support Bay",
    "Protoss Temple",
    "Fleet Beacon",
    "Explosion (Large)",
    "Protoss Building Rubble (Small)",
    "Protoss Building Rubble (Large)",
    "Battlecruiser",
    "Civilian",
    "Dropship",
    "Firebat",
    "Ghost",
    "Ghost Remnants",
    "Nuke Target Dot",
    "Goliath Base",
    "Goliath Turret",
    "Sarah Kerrigan",
    "Marine",
    "Marine Remnants",
    "Scanner Sweep",
    "Wraith",
    "SCV",
    "Siege Tank (Tank) Base",
    "Siege Tank (Tank) Turret",
    "Siege Tank (Siege) Base",
    "Siege Tank (Siege) Turret",
    "Vulture",
    "Spider Mine",
    "Science Vessel (Base)",
    "Science Vessel (Turret)",
    "Terran Academy",
    "Barracks",
    "Armory",
    "Comsat Station",
    "Command Center",
    "Supply Depot",
    "Control Tower",
    "Factory",
    "Covert Ops",
    "Ion Cannon",
    "Machine Shop",
    "Missile Turret (Base)",
    "Crashed Batlecruiser",
    "Physics Lab",
    "Bunker",
    "Refinery",
    "Science Facility",
    "Nuclear Silo",
    "Nuclear Missile",
    "Nuke Hit",
    "Starport",
    "Engineering Bay",
    "Terran Construction (Large)",
    "Terran Construction (Small)",
    "Building Explosion (Large)",
    "Terran Building Rubble (Small)",
    "Terran Building Rubble (Large)",
    "Vespene Geyser",
    "Ragnasaur (Ash)",
    "Rhynadon (Badlands)",
    "Bengalaas (Jungle)",
    "Mineral Field Type1",
    "Mineral Field Type2",
    "Mineral Field Type3",
    "Independent Starport (Unused)",
    "Zerg Beacon",
    "Terran Beacon",
    "Protoss Beacon",
    "Dark Swarm",
    "Flag",
    "Young Chrysalis",
    "Psi Emitter",
    "Data Disc",
    "Khaydarin Crystal",
    "Mineral Chunk Type1",
    "Mineral Chunk Type2",
    "Protoss Gas Orb Type1",
    "Protoss Gas Orb Type2",
    "Zerg Gas Sac Type1",
    "Zerg Gas Sac Type2",
    "Terran Gas Tank Type1",
    "Terran Gas Tank Type2",
    "White Circle (Invisible)",
    "Start Location",
    "Map Revealer",
    "Floor Gun Trap",
    "Wall Missile Trap",
    "Wall Missile Trap2",
    "Wall Flame Trap",
    "Wall Flame Trap2",
    "Floor Missile Trap",
    "Longbolt/Gemini Missiles Trail",
    "Grenade Shot Smoke",
    "Vespene Geyser Smoke1",
    "Vespene Geyser Smoke2",
    "Vespene Geyser Smoke3",
    "Vespene Geyser Smoke4",
    "Vespene Geyser Smoke5",
    "Small Explosion (Unused)",
    "Double Explosion",
    "Cursor Marker",
    "Egg Spawn",
    "High Templar Glow",
    "Psi Field (Right Upper)",
    "Burrowing Dust",
    "Building Landing Dust Type1",
    "Building Landing Dust Type2",
    "Building Landing Dust Type3",
    "Building Landing Dust Type4",
    "Building Landing Dust Type5",
    "Building Lifting Dust Type1",
    "Building Lifting Dust Type2",
    "Building Lifting Dust Type3",
    "Building Lifting Dust Type4",
    "Needle Spines",
    "Dual Photon Blasters Hit",
    "Particle Beam Hit",
    "Anti-Matter Missile",
    "Pulse Cannon",
    "Phase Disruptor",
    "STA/STS Photon Cannon Overlay",
    "Psionic Storm",
    "Fusion Cutter Hit",
    "Gauss Rifle Hit",
    "Gemini Missiles",
    "Fragmentation Grenade",
    "Magna Pulse (Unused)",
    "Lockdown/LongBolt/Hellfire Missile",
    "C-10 Canister Rifle Hit",
    "ATS/ATA Laser Battery",
    "Burst Lasers",
    "Arclite Shock Cannon Hit",
    "Yamato Gun",
    "Yamato Gun Trail",
    "EMP Shockwave Missile",
    "Needle Spine Hit",
    "Plasma Drip Hit (Unused)",
    "Sunken Colony Tentacle",
    "Venom (Unused Zerg Weapon)",
    "Acid Spore",
    "Glave Wurm",
    "Seeker Spores",
    "Queen Spell Holder",
    "Stasis Field Hit",
    "Plague Cloud",
    "Consume",
    "Ensnare",
    "Glave Wurm/Seeker Spores Hit",
    "Psionic Shockwave Hit",
    "Glave Wurm Trail",
    "Seeker Spores Overlay",
    "Phase Disruptor (Unused)",
    "White Circle",
    "Acid Spray (Unused)",
    "Plasma Drip (Unused)",
    "Scarab/Anti-Matter Missile Overlay",
    "Hallucination Death1",
    "Hallucination Death2",
    "Hallucination Death3",
    "Bunker Overlay",
    "FlameThrower",
    "Recall Field",
    "Scanner Sweep Hit",
    "Left Upper Level Door",
    "Right Upper Level Door",
    "Substructure Left Door",
    "Substructure Right Door",
    "Substructure Opening Hole",
    "7/13 Twilight",
    "7/14 Twilight",
    "7/16 Twilight",
    "7/15 Twilight",
    "7/19 Twilight",
    "7/20 Twilight",
    "7/21 Twilight",
    "Unknown Twilight",
    "7/17 Twilight",
    "6/1 Twilight",
    "6/2 Twilight",
    "6/3 Twilight",
    "6/4 Twilight",
    "6/5 Twilight",
    "8/3 Twilight",
    "8/4 Twilight",
    "9/29 Ice",
    "9/28 Ice",
    "12/38 Ice",
    "12/37 Ice",
    "12/33 Ice",
    "9/21 Ice",
    "9/15 Ice",
    "9/16 Ice",
    "Unknown410",
    "Unknown411",
    "12/9 Ice1",
    "12/10 Ice",
    "9/24 Ice",
    "9/23 Ice",
    "Unknown416",
    "12/7 Ice",
    "12/8 Ice",
    "12/9 Ice2",
    "12/11 Ice",
    "12/40 Ice",
    "12/41 Ice",
    "12/42 Ice",
    "12/5 Ice",
    "12/6 Ice",
    "12/36 Ice",
    "12/32 Ice",
    "12/35 Ice",
    "12/34 Ice",
    "12/24 Ice",
    "12/25 Ice",
    "9/22 Ice",
    "12/31 Ice",
    "12/20 Ice",
    "12/26 Ice",
    "12/27 Ice",
    "12/30 Ice",
    "12/39 Ice",
    "Unknown439",
    "4/1 Ice",
    "6/1 Ice",
    "5/6 Ice",
    "5/7 Ice",
    "5/8 Ice",
    "5/9 Ice",
    "10/10 Desert1",
    "10/12 Desert1",
    "10/8 Desert",
    "10/9 Desert1",
    "6/10 Desert",
    "6/13 Desert",
    "Unknown Desert",
    "10/12 Desert2",
    "10/9 Desert2",
    "10/10 Desert2",
    "10/11 Desert",
    "10/14 Desert",
    "10/41 Desert",
    "10/39 Desert",
    "10/9 Desert",
    "10/6 Desert",
    "10/7 Desert",
    "4/6 Desert",
    "4/11 Desert",
    "4/10 Desert",
    "4/9 Desert",
    "4/7 Desert",
    "4/12 Desert",
    "4/8 Desert",
    "4/13 Desert",
    "6/14 Desert",
    "4/17 Desert",
    "6/20 Desert",
    "4/15 Desert1",
    "4/15 Desert2",
    "10/23 Desert",
    "8/23 Desert",
    "10/5 Desert",
    "12/1 Desert Overlay",
    "11/3 Desert",
    "Lurker Egg",
    "Devourer",
    "Devourer Death",
    "Lurker Remnants",
    "Lurker",
    "Dark Archon Energy",
    "Corsair",
    "Dark Templar (Unit)",
    "Medic",
    "Medic Remnants",
    "Valkyrie",
    "Scantid (Desert)",
    "Kakaru (Twilight)",
    "Ursadon (Ice)",
    "Overmind Cocoon",
    "Power Generator",
    "Xel'Naga Temple",
    "Psi Disrupter",
    "Warp Gate",
    "Feedback Hit (Small)",
    "Feedback Hit (Medium)",
    "Feedback Hit (Large)",
    "Disruption Web",
    "White Circle2",
    "Halo Rockets Trail",
    "Neutron Flare",
    "Neutron Flare Overlay (Unused)",
    "Optical Flare Grenade",
    "Halo Rockets",
    "Subterranean Spines Hit",
    "Subterranean Spines",
    "Corrosive Acid Shot",
    "Corrosive Acid Hit",
    "Maelstrom Hit",
    "Uraj",
    "Khalis",
]



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from typing_extensions import Self

from ...localize import _
from ...utils import EPError
from ..allocator.payload import _PayloadBuffer
from .eudobj import EUDObject


class Db(EUDObject):
    """Class for raw data object"""

    def __new__(cls, *args, **kwargs) -> Self:
        return super().__new__(cls)

    def __init__(self, b: bytes | int | str, /) -> None:
        super().__init__()
        if isinstance(b, str):
            b = b.encode("UTF-8")
            if 0 in b:
                raise EPError(
                    _("no nul bytes allowed in the middle of {}").format(
                        self.__class__
                    )
                )
            b += b"\0"
        self.content: bytes = bytes(b)

    def GetDataSize(self) -> int:  # noqa: N802
        return len(self.content)

    def WritePayload(self, pbuffer: _PayloadBuffer) -> None:  # noqa: N802
        pbuffer.WriteBytes(self.content)



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from ... import utils as ut
from ...localize import _
from .chktok import CHK
from .unitprp import UnitProperty, property_key

_uprpdict: dict[bytes, int] = {}
_uprptable: list[bytes] = []


def init_property_map(chkt: CHK) -> None:
    global _prptable, _uprpdict
    _uprpdict.clear()
    _uprptable.clear()
    _uprptable.extend([b""] * 64)

    # Backup UPRP data
    uprp = chkt.getsection("UPRP")
    try:
        upus = chkt.getsection("UPUS")
    except KeyError:
        upus = bytes([1] * 64)

    for i in range(64):
        if upus[i]:
            uprpdata = property_key(uprp, 20 * i)
            # FIXME: merge duplicates and fix existing CUWPs
            _uprpdict[uprpdata] = i
            _uprptable[i] = uprpdata


def GetPropertyIndex(prop: UnitProperty | bytes) -> int:  # noqa: N802
    ut.ep_assert(
        isinstance(prop, UnitProperty) or isinstance(prop, bytes),
        _("Invalid property type"),
    )

    prop = property_key(bytes(prop))
    try:
        return _uprpdict[prop] + 1  # SC counts unit properties from 1.

    except KeyError:
        index_found = False
        for uprpindex in range(64):
            if not _uprptable[uprpindex]:
                index_found = True
                break

        ut.ep_assert(index_found, _("Unit property table overflow"))

        _uprptable[uprpindex] = prop
        _uprpdict[prop] = uprpindex
        return uprpindex + 1  # SC counts unit properties from 1.


def apply_property_map(chkt: CHK) -> None:
    uprpdata = b"".join([uprpdata or bytes(20) for uprpdata in _uprptable])
    chkt.setsection("UPRP", uprpdata)



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.
from typing import Any

from ..localize import _
from ..utils import ep_warn

_objns: dict[str, Any] = {}


def EUDClearNamespace() -> None:  # noqa: N802
    _objns.clear()


def EUDRegisterObjectToNamespace(funcname: str, obj: Any) -> Any:  # noqa: N802
    """Register object to inline code namespace."""
    if funcname[0] != "_":
        if funcname in _objns:
            err = _("Duplicated name {} for EUDRegisterObjectToNamespace")
            ep_warn(err.format(funcname))
            _objns[funcname] = None
        else:
            _objns[funcname] = obj

    return obj


def EUDRegistered(func):  # noqa: N802
    """Decorator for registering class / function."""
    return EUDRegisterObjectToNamespace(func.__name__, func)


def GetEUDNamespace() -> dict[str, Any]:  # noqa: N802
    """Get list of functions that inline code can use."""
    return _objns



# Copyright 2023 by Armoha.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

# ruff: noqa: N815
from typing import ClassVar, Self, cast

from .. import core as c
from .. import ctrlstru as cs
from ..core.eudfunc.eudf import _EUDPredefineReturn
from ..localize import _
from ..utils import EPD, EPError, classproperty, unProxy
from .csprite import int_or_var
from .offsetmap import (
    BoolMember,
    ButtonSetMember,
    ByteEnumMember,
    ByteMember,
    CSpriteMember,
    CUnitMember,
    DwordEnumMember,
    DwordMember,
    EPDOffsetMap,
    Flag,
    FlingyMember,
    MovementFlags,
    PlayerMember,
    PositionMember,
    PositionXMember,
    PositionYMember,
    TechMember,
    UnitMember,
    UnitOrderMember,
    UnsupportedMember,
    UpgradeMember,
    WordMember,
    WorkerCarryTypeMember,
)
from .offsetmap.epdoffsetmap import _epd_cache, _ptr_cache
from .player import CurrentPlayer, TrgPlayer
from .unit import TrgUnit


class StatusFlags(DwordEnumMember):
    __slots__ = ()
    Completed = Flag(0x00000001)
    GroundedBuilding = Flag(0x00000002)
    "a building that is on the ground"
    InAir = Flag(0x00000004)
    Disabled = Flag(0x00000008)
    "Protoss Unpowered"
    Burrowed = Flag(0x00000010)
    InBuilding = Flag(0x00000020)
    InTransport = Flag(0x00000040)
    CanBeChased = Flag(0x00000080)
    RequiresDetection = Flag(0x00000100)
    Cloaked = Flag(0x00000200)
    DoodadStatesThing = Flag(0x00000400)
    "protoss unpowered buildings have this flag set"
    CloakingForFree = Flag(0x00000800)
    "Requires no energy to cloak"
    CanNotReceiveOrders = Flag(0x00001000)
    NoBrkCodeStart = Flag(0x00002000)
    "Unbreakable code section in iscript"
    UNKNOWN2 = Flag(0x00004000)
    CanNotAttack = Flag(0x00008000)
    CanTurnAroundToAttack = Flag(0x00010000)
    "canAttack? named IsAUnit in BWAPI"
    IsBuilding = Flag(0x00020000)
    IgnoreTileCollision = Flag(0x00040000)
    Unmovable = Flag(0x00080000)
    IsNormal = Flag(0x00100000)
    "0 for ignore terrain elevation while moving"
    NoCollide = Flag(0x00200000)
    "if set, other units wont collide with the unit (like burrowed units)"
    UNKNOWN5 = Flag(0x00400000)
    IsGathering = Flag(0x00800000)
    "if set, the unit wont collide with other units (like workers gathering)"
    SubunitWalking = Flag(0x01000000)
    "makes subunit play walking animation if base unit movement flag 2 is set"
    SubunitFollow = Flag(0x02000000)
    "makes subunit instantly follow base unit's facing direction. Turret related"
    Invincible = Flag(0x04000000)
    HoldingPosition = Flag(0x08000000)
    "Set if the unit is currently holding position"
    SpeedUpgrade = Flag(0x10000000)
    CooldownUpgrade = Flag(0x20000000)
    IsHallucination = Flag(0x40000000)
    "1 for hallucinated units, 0 for 'normal' units"
    IsSelfDestructing = Flag(0x80000000)
    "Set for when the unit is self-destructing (scarab, scourge, infested terran)"


class PathingFlags(ByteEnumMember):
    __slots__ = ()
    HasCollision = Flag(0x01)
    IsStacked = Flag(0x02)
    Decollide = Flag(0x04)


class CUnit(EPDOffsetMap):
    __slots__ = "_ptr"
    prev: ClassVar[CUnitMember] = CUnitMember("struct", 0x000)
    next: ClassVar[CUnitMember] = CUnitMember("struct", 0x004)
    hp: ClassVar[DwordMember] = DwordMember("struct", 0x008)
    "displayed value is ceil(healthPoints/256)"
    sprite: ClassVar[CSpriteMember] = CSpriteMember("struct", 0x00C)
    moveTargetPos: ClassVar[PositionMember] = PositionMember("struct", 0x010)
    moveTargetX: ClassVar[PositionXMember] = PositionXMember("struct", 0x010)
    moveTargetY: ClassVar[PositionYMember] = PositionYMember("struct", 0x012)
    moveTarget: ClassVar[CUnitMember] = CUnitMember("struct", 0x014)
    moveTargetUnit = moveTarget
    nextMovementWaypoint: ClassVar[PositionMember] = PositionMember("struct", 0x018)
    """The next way point in the path the unit is following to get to its
    destination. Equal to moveToPos for air units since they don't need to
    navigate around buildings."""
    nextTargetWaypoint: ClassVar[PositionMember] = PositionMember("struct", 0x01C)
    "The desired position"
    movementFlags = MovementFlags("struct", 0x020)
    currentDirection1: ClassVar[ByteMember] = ByteMember("struct", 0x021)
    "current direction the unit is facing"
    turnSpeed: ClassVar[ByteMember] = ByteMember("struct", 0x022)  # flingy
    turnRadius = turnSpeed
    velocityDirection1: ClassVar[ByteMember] = ByteMember("struct", 0x023)
    """usually only differs from the currentDirection field for units that
    can accelerate and travel in a different direction than they are facing.
    For example Mutalisks can change the direction they are facing
    faster than then can change the direction they are moving."""
    flingyID: ClassVar[FlingyMember] = FlingyMember("struct", 0x024)
    unknown0x26: ClassVar[ByteMember] = ByteMember("struct", 0x026)
    flingyMovementType: ClassVar[ByteMember] = ByteMember("struct", 0x027)
    pos: ClassVar[PositionMember] = PositionMember("struct", 0x028)
    "Current position of the unit"
    posX: ClassVar[PositionXMember] = PositionXMember("struct", 0x028)
    posY: ClassVar[PositionYMember] = PositionYMember("struct", 0x02A)
    haltX: ClassVar[DwordMember] = DwordMember("struct", 0x02C)
    haltY: ClassVar[DwordMember] = DwordMember("struct", 0x030)
    topSpeed: ClassVar[DwordMember] = DwordMember("struct", 0x034)
    currentSpeed1: ClassVar[DwordMember] = DwordMember("struct", 0x038)
    currentSpeed2: ClassVar[DwordMember] = DwordMember("struct", 0x03C)
    currentVelocityX: ClassVar[DwordMember] = DwordMember("struct", 0x040)
    currentVelocityY: ClassVar[DwordMember] = DwordMember("struct", 0x044)
    acceleration: ClassVar[WordMember] = WordMember("struct", 0x048)
    currentDirection2: ClassVar[ByteMember] = ByteMember("struct", 0x04A)
    velocityDirection2: ClassVar[ByteMember] = ByteMember("struct", 0x04B)
    "pathing related"
    owner: ClassVar[PlayerMember] = PlayerMember("struct", 0x04C)
    playerID = owner
    order: ClassVar[UnitOrderMember] = UnitOrderMember("struct", 0x04D)
    orderID: ClassVar[UnitOrderMember] = UnitOrderMember("struct", 0x04D)
    orderState: ClassVar[ByteMember] = ByteMember("struct", 0x04E)
    orderSignal: ClassVar[ByteMember] = ByteMember("struct", 0x04F)
    orderUnitType: ClassVar[UnitMember] = UnitMember("struct", 0x050)
    unknown0x52: ClassVar[WordMember] = WordMember("struct", 0x052)
    "2-byte padding"
    cooldown: ClassVar[DwordMember] = DwordMember("struct", 0x054)
    orderTimer: ClassVar[ByteMember] = ByteMember("struct", 0x054)
    groundWeaponCooldown: ClassVar[ByteMember] = ByteMember("struct", 0x055)
    gCooldown = groundWeaponCooldown
    airWeaponCooldown: ClassVar[ByteMember] = ByteMember("struct", 0x056)
    aCooldown = airWeaponCooldown
    spellCooldown: ClassVar[ByteMember] = ByteMember("struct", 0x057)
    # ActionFocus
    orderTargetXY: ClassVar[PositionMember] = PositionMember("struct", 0x058)
    orderTargetPos = orderTargetXY
    orderTargetX: ClassVar[PositionXMember] = PositionXMember("struct", 0x058)
    orderTargetY: ClassVar[PositionYMember] = PositionYMember("struct", 0x05A)
    orderTarget: ClassVar[CUnitMember] = CUnitMember("struct", 0x05C)
    orderTargetUnit: ClassVar[CUnitMember] = CUnitMember("struct", 0x05C)
    shield: ClassVar[DwordMember] = DwordMember("struct", 0x060)
    unitType: ClassVar[UnitMember] = UnitMember("struct", 0x064)
    unitID = unitType
    unknown0x66: ClassVar[WordMember] = WordMember("struct", 0x066)
    "2-byte padding"
    prevPlayerUnit: ClassVar[CUnitMember] = CUnitMember("struct", 0x068)
    nextPlayerUnit: ClassVar[CUnitMember] = CUnitMember("struct", 0x06C)
    subUnit: ClassVar[CUnitMember] = CUnitMember("struct", 0x070)
    orderQueueHead: ClassVar[UnsupportedMember] = UnsupportedMember("struct", 0x074)
    """COrder"""
    orderQueueTail: ClassVar[UnsupportedMember] = UnsupportedMember("struct", 0x078)
    """COrder"""
    autoTargetUnit: ClassVar[CUnitMember] = CUnitMember("struct", 0x07C)
    connectedUnit: ClassVar[CUnitMember] = CUnitMember("struct", 0x080)
    "larva, in-transit, addons"
    orderQueueCount: ClassVar[ByteMember] = ByteMember("struct", 0x084)
    "may be count in addition to first since can be 2 when 3 orders are queued"
    orderQueueTimer: ClassVar[ByteMember] = ByteMember("struct", 0x085)
    "Cycles down from from 8 to 0 (inclusive). See also 0x122."
    unknown0x86: ClassVar[ByteMember] = ByteMember("struct", 0x086)
    attackNotifyTimer: ClassVar[ByteMember] = ByteMember("struct", 0x087)
    "Prevent 'Your forces are under attack.' on every attack"
    prevUnitType: ClassVar[UnsupportedMember] = UnsupportedMember("struct", 0x088)
    "Unit. Zerg buildings while morphing"
    lastEventTimer: ClassVar[UnsupportedMember] = UnsupportedMember("struct", 0x08A)
    """Byte"""
    lastEventColor: ClassVar[UnsupportedMember] = UnsupportedMember("struct", 0x08B)
    "Byte. 17 = was completed (train, morph), 174 = was attacked"
    unknown0x8C: ClassVar[WordMember] = WordMember("struct", 0x08C)
    "might have originally been RGB from lastEventColor"
    rankIncrease: ClassVar[ByteMember] = ByteMember("struct", 0x08E)
    killCount: ClassVar[ByteMember] = ByteMember("struct", 0x08F)
    lastAttackingPlayer: ClassVar[PlayerMember] = PlayerMember("struct", 0x090)
    secondaryOrderTimer: ClassVar[ByteMember] = ByteMember("struct", 0x091)
    AIActionFlag: ClassVar[ByteMember] = ByteMember("struct", 0x092)
    userActionFlags: ClassVar[ByteMember] = ByteMember("struct", 0x093)
    """2 = issued an order,
    3 = interrupted an order,
    4 = hide self before death (self-destruct?)"""
    currentButtonSet: ClassVar[ButtonSetMember] = ButtonSetMember("struct", 0x094)
    isCloaked: ClassVar[BoolMember] = BoolMember("struct", 0x096)
    movementState: ClassVar[ByteMember] = ByteMember("struct", 0x097)
    buildQueue1: ClassVar[UnitMember] = UnitMember("struct", 0x098)
    buildQueue2: ClassVar[UnitMember] = UnitMember("struct", 0x09A)
    buildQueue3: ClassVar[UnitMember] = UnitMember("struct", 0x09C)
    buildQueue4: ClassVar[UnitMember] = UnitMember("struct", 0x09E)
    buildQueue5: ClassVar[UnitMember] = UnitMember("struct", 0x0A0)
    buildQueue12: ClassVar[DwordMember] = DwordMember("struct", 0x098)
    buildQueue34: ClassVar[DwordMember] = DwordMember("struct", 0x09C)
    energy: ClassVar[WordMember] = WordMember("struct", 0x0A2)
    buildQueueSlot: ClassVar[ByteMember] = ByteMember("struct", 0x0A4)
    uniquenessIdentifier: ClassVar[ByteMember] = ByteMember("struct", 0x0A5)
    targetOrderSpecial = uniquenessIdentifier
    secondaryOrder: ClassVar[UnitOrderMember] = UnitOrderMember("struct", 0x0A6)
    secondaryOrderID = secondaryOrder
    buildingOverlayState: ClassVar[ByteMember] = ByteMember("struct", 0x0A7)
    "0 means the building has the largest amount of fire/blood"
    hpGain: ClassVar[WordMember] = WordMember("struct", 0x0A8)
    "buildRepairHpGain"
    shieldGain: ClassVar[WordMember] = WordMember("struct", 0x0AA)
    "Shield gain on construction"
    remainingBuildTime: ClassVar[WordMember] = WordMember("struct", 0x0AC)
    """Remaining bulding time; also used by powerups (flags)
    as the timer for returning to their original location."""
    prevHp: ClassVar[WordMember] = WordMember("struct", 0x0AE)
    """The HP of the unit before it changed
    (example: Drone->Hatchery, the Drone's HP will be stored here)"""
    loadedUnit1: ClassVar[UnsupportedMember] = UnsupportedMember("struct", 0x0B0)
    "Word. alphaID (StoredUnit)"
    loadedUnit2: ClassVar[UnsupportedMember] = UnsupportedMember("struct", 0x0B2)
    "Word. alphaID (StoredUnit)"
    loadedUnit3: ClassVar[UnsupportedMember] = UnsupportedMember("struct", 0x0B4)
    "Word. alphaID (StoredUnit)"
    loadedUnit4: ClassVar[UnsupportedMember] = UnsupportedMember("struct", 0x0B6)
    "Word. alphaID (StoredUnit)"
    loadedUnit5: ClassVar[UnsupportedMember] = UnsupportedMember("struct", 0x0B8)
    "Word. alphaID (StoredUnit)"
    loadedUnit6: ClassVar[UnsupportedMember] = UnsupportedMember("struct", 0x0BA)
    "Word. alphaID (StoredUnit)"
    loadedUnit7: ClassVar[UnsupportedMember] = UnsupportedMember("struct", 0x0BC)
    "Word. alphaID (StoredUnit)"
    loadedUnit8: ClassVar[UnsupportedMember] = UnsupportedMember("struct", 0x0BE)
    "Word. alphaID (StoredUnit)"
    # union (0xC0 ~ 0xCF) //==================================
    spiderMineCount: ClassVar[ByteMember] = ByteMember("struct", 0x0C0)  # vulture
    # carrier, reaver ----------------------------------------
    inHangarChild: ClassVar[CUnitMember] = CUnitMember("struct", 0x0C0)
    outHangarChild: ClassVar[CUnitMember] = CUnitMember("struct", 0x0C4)
    inHangarCount: ClassVar[ByteMember] = ByteMember("struct", 0x0C8)
    outHangarCount: ClassVar[ByteMember] = ByteMember("struct", 0x0C9)
    # interceptor, scarab ------------------------------------
    parent: ClassVar[CUnitMember] = CUnitMember("struct", 0x0C0)
    prevFighter: ClassVar[CUnitMember] = CUnitMember("struct", 0x0C4)
    nextFighter: ClassVar[CUnitMember] = CUnitMember("struct", 0x0C8)
    isOutsideHangar: ClassVar[BoolMember] = BoolMember("struct", 0x0CC)
    # beacon -------------------------------------------------
    beaconUnknown0xC0: ClassVar[DwordMember] = DwordMember("struct", 0x0C0)
    beaconUnknown0xC4: ClassVar[DwordMember] = DwordMember("struct", 0x0C4)
    flagSpawnFrame: ClassVar[DwordMember] = DwordMember("struct", 0x0C8)  # beacon
    # building /==============================================
    addon: ClassVar[CUnitMember] = CUnitMember("struct", 0x0C0)
    addonBuildType: ClassVar[UnitMember] = UnitMember("struct", 0x0C4)
    upgradeResearchTime: ClassVar[WordMember] = WordMember("struct", 0x0C6)
    techType: ClassVar[TechMember] = TechMember("struct", 0x0C8)
    upgradeType: ClassVar[UpgradeMember] = UpgradeMember("struct", 0x0C9)
    larvaTimer: ClassVar[ByteMember] = ByteMember("struct", 0x0CA)
    landingTimer: ClassVar[ByteMember] = ByteMember("struct", 0x0CB)
    creepTimer: ClassVar[ByteMember] = ByteMember("struct", 0x0CC)
    upgradeLevel: ClassVar[ByteMember] = ByteMember("struct", 0x0CD)
    # padding0xCE
    # resource -----------------------------------------------
    resourceAmount: ClassVar[WordMember] = WordMember("struct", 0x0D0)  # 0x0D0 union
    resourceIscript: ClassVar[ByteMember] = ByteMember("struct", 0x0D2)
    gatherQueueCount: ClassVar[BoolMember] = BoolMember("struct", 0x0D3)
    """it is byte but effectively bool; always set to 1 when beginning to harvest,
    but when finshed, it is    by 1 instead of set to 0"""
    nextGatherer: ClassVar[CUnitMember] = CUnitMember("struct", 0x0D4)
    "pointer to the next worker unit waiting in line to gather"
    resourceGroup: ClassVar[ByteMember] = ByteMember("struct", 0x0D8)
    resourceBelongsToAI: ClassVar[BoolMember] = BoolMember("struct", 0x0D9)
    # other buildings ----------------------------------------
    nydusExit: ClassVar[CUnitMember] = CUnitMember("struct", 0x0D0)
    "connected nydus canal"
    # FIXME: should be CThingy
    ghostNukeDot: ClassVar[DwordMember] = DwordMember("struct", 0x0D0)
    """CThingy struct is same as CUnit but trimmed down to [prev, next, hp, sprite],
    with "hp" field used as unitID for fog thingies or otherwise unused"""
    pylonAura: ClassVar[CSpriteMember] = CSpriteMember("struct", 0x0D0)
    # silo
    siloNuke: ClassVar[CUnitMember] = CUnitMember("struct", 0x0D0)
    siloReady: ClassVar[BoolMember] = BoolMember("struct", 0x0D4)
    # hatchery
    hatcheryHarvestLT: ClassVar[DwordMember] = DwordMember("struct", 0x0D0)
    hatcheryHarvestRB: ClassVar[DwordMember] = DwordMember("struct", 0x0D4)
    hatcheryHarvestL: ClassVar[WordMember] = WordMember("struct", 0x0D0)
    hatcheryHarvestT: ClassVar[WordMember] = WordMember("struct", 0x0D2)
    hatcheryHarvestR: ClassVar[WordMember] = WordMember("struct", 0x0D4)
    hatcheryHarvestB: ClassVar[WordMember] = WordMember("struct", 0x0D6)
    # ==============================================/ building
    # worker -------------------------------------------------
    powerup: ClassVar[CUnitMember] = CUnitMember("struct", 0x0C0)
    targetResourcePos: ClassVar[PositionMember] = PositionMember("struct", 0x0C4)
    targetResourceX: ClassVar[PositionXMember] = PositionXMember("struct", 0x0C4)
    targetResourceY: ClassVar[PositionYMember] = PositionYMember("struct", 0x0C6)
    targetResourceUnit: ClassVar[CUnitMember] = CUnitMember("struct", 0x0C8)
    repairResourceLossTimer: ClassVar[WordMember] = WordMember("struct", 0x0CC)
    isCarryingSomething: ClassVar[BoolMember] = BoolMember("struct", 0x0CE)
    resourceCarryAmount: ClassVar[ByteMember] = ByteMember("struct", 0x0CF)
    harvestTarget: ClassVar[CUnitMember] = CUnitMember("struct", 0x0D0)
    prevHarvestUnit: ClassVar[CUnitMember] = CUnitMember("struct", 0x0D4)
    nextHarvestUnit: ClassVar[CUnitMember] = CUnitMember("struct", 0x0D8)
    "When there is a gather conflict"
    # powerup ------------------------------------------------
    powerupOrigin: ClassVar[PositionMember] = PositionMember("struct", 0x0D0)
    powerupOriginX: ClassVar[PositionXMember] = PositionXMember("struct", 0x0D0)
    powerupOriginY: ClassVar[PositionYMember] = PositionYMember("struct", 0x0D2)
    powerupCarryingUnit: ClassVar[CUnitMember] = CUnitMember("struct", 0x0D4)
    # \\\\\\\\\\\\\\\=================================// union
    statusFlags = StatusFlags("struct", 0x0DC)
    resourceType: ClassVar[WorkerCarryTypeMember] = WorkerCarryTypeMember(
        "struct", 0x0E0
    )
    "Type of resource chunk carried by worker: (None, Gas, Ore, GasOrOre, PowerUp)"
    wireframeRandomizer: ClassVar[ByteMember] = ByteMember("struct", 0x0E1)
    secondaryOrderState: ClassVar[ByteMember] = ByteMember("struct", 0x0E2)
    recentOrderTimer: ClassVar[ByteMember] = ByteMember("struct", 0x0E3)
    """Counts down from 15 to 0 when most orders are given,
    or when the unit moves after reaching a patrol location"""
    visibilityStatus: ClassVar[DwordMember] = DwordMember("struct", 0x0E4)
    "which players can detect this unit (cloaked/burrowed)"
    secondaryOrderPos: ClassVar[PositionMember] = PositionMember("struct", 0x0E8)
    secondaryOrderX: ClassVar[PositionXMember] = PositionXMember("struct", 0x0E8)
    secondaryOrderY: ClassVar[PositionYMember] = PositionYMember("struct", 0x0EA)
    currentBuildUnit: ClassVar[CUnitMember] = CUnitMember("struct", 0x0EC)
    prevBurrowedUnit: ClassVar[UnsupportedMember] = UnsupportedMember(
        "struct", 0x0F0
    )  # CUnit
    nextBurrowedUnit: ClassVar[UnsupportedMember] = UnsupportedMember(
        "struct", 0x0F4
    )  # CUnit
    rallyPos: ClassVar[PositionMember] = PositionMember("struct", 0x0F8)
    rallyX: ClassVar[PositionXMember] = PositionXMember("struct", 0x0F8)
    rallyY: ClassVar[PositionYMember] = PositionYMember("struct", 0x0FA)
    rallyUnit: ClassVar[CUnitMember] = CUnitMember("struct", 0x0FC)
    prevPsiProvider: ClassVar[CUnitMember] = CUnitMember("struct", 0x0F8)
    nextPsiProvider: ClassVar[CUnitMember] = CUnitMember("struct", 0x0FC)
    path: ClassVar[UnsupportedMember] = UnsupportedMember("struct", 0x100)  # Dword
    pathingCollisionInterval: ClassVar[ByteMember] = ByteMember("struct", 0x104)
    pathingFlags = PathingFlags("struct", 0x105)
    unknown0x106: ClassVar[ByteMember] = ByteMember("struct", 0x106)
    isBeingHealed: ClassVar[BoolMember] = BoolMember("struct", 0x107)
    "1 if a medic is currently healing this unit"
    contourBoundsLT: ClassVar[UnsupportedMember] = UnsupportedMember("struct", 0x108)
    """Dword"""
    "A rect that specifies the closest contour (collision) points"
    contourBoundsRB: ClassVar[UnsupportedMember] = UnsupportedMember("struct", 0x10C)
    """Dword"""
    contourBoundsL: ClassVar[UnsupportedMember] = UnsupportedMember("struct", 0x108)
    """Word"""
    contourBoundsT: ClassVar[UnsupportedMember] = UnsupportedMember("struct", 0x10A)
    """Word"""
    contourBoundsR: ClassVar[UnsupportedMember] = UnsupportedMember("struct", 0x10C)
    """Word"""
    contourBoundsB: ClassVar[UnsupportedMember] = UnsupportedMember("struct", 0x10E)
    """Word"""
    removeTimer: ClassVar[WordMember] = WordMember("struct", 0x110)
    """Hallucination, Dark Swarm, Disruption Web, Broodling
    (but not Scanner Sweep according to BWAPI)"""
    defensiveMatrixHp: ClassVar[WordMember] = WordMember("struct", 0x112)
    defensiveMatrixTimer: ClassVar[ByteMember] = ByteMember("struct", 0x114)
    stimTimer: ClassVar[ByteMember] = ByteMember("struct", 0x115)
    ensnareTimer: ClassVar[ByteMember] = ByteMember("struct", 0x116)
    lockdownTimer: ClassVar[ByteMember] = ByteMember("struct", 0x117)
    irradiateTimer: ClassVar[ByteMember] = ByteMember("struct", 0x118)
    stasisTimer: ClassVar[ByteMember] = ByteMember("struct", 0x119)
    plagueTimer: ClassVar[ByteMember] = ByteMember("struct", 0x11A)
    stormTimer: ClassVar[ByteMember] = ByteMember("struct", 0x11B)
    isUnderStorm: ClassVar[BoolMember] = BoolMember("struct", 0x11B)
    "Used to tell if a unit is under psi storm"
    irradiatedBy: ClassVar[CUnitMember] = CUnitMember("struct", 0x11C)
    irradiatePlayerID: ClassVar[PlayerMember] = PlayerMember("struct", 0x120)
    parasiteFlags: ClassVar[ByteMember] = ByteMember("struct", 0x121)
    "Each bit corresponds to the player who has parasited this unit"
    cycleCounter: ClassVar[ByteMember] = ByteMember("struct", 0x122)
    "counts/cycles up from 0 to 7 (inclusive). See also 0x85"
    blindFlags: ClassVar[ByteMember] = ByteMember("struct", 0x123)
    """Each bit corresponds to the player who has optical flared this unit,
    like parasiteFlags, but is read as a bool for vision check"""
    maelstromTimer: ClassVar[ByteMember] = ByteMember("struct", 0x124)
    unusedTimer: ClassVar[ByteMember] = ByteMember("struct", 0x125)
    "Might be afterburner timer or ultralisk roar timer"
    acidSporeCount: ClassVar[ByteMember] = ByteMember("struct", 0x126)
    acidSporeTime0: ClassVar[ByteMember] = ByteMember("struct", 0x127)
    acidSporeTime1: ClassVar[ByteMember] = ByteMember("struct", 0x128)
    acidSporeTime2: ClassVar[ByteMember] = ByteMember("struct", 0x129)
    acidSporeTime3: ClassVar[ByteMember] = ByteMember("struct", 0x12A)
    acidSporeTime4: ClassVar[ByteMember] = ByteMember("struct", 0x12B)
    acidSporeTime5: ClassVar[ByteMember] = ByteMember("struct", 0x12C)
    acidSporeTime6: ClassVar[ByteMember] = ByteMember("struct", 0x12D)
    acidSporeTime7: ClassVar[ByteMember] = ByteMember("struct", 0x12E)
    acidSporeTime8: ClassVar[ByteMember] = ByteMember("struct", 0x12F)
    offsetIndex3by3: ClassVar[UnsupportedMember] = UnsupportedMember("struct", 0x130)
    """Word"""
    """Cycles between 0-12 for each bullet fired by this unit
    (if it uses a 'Attack 3x3 area' weapon)"""
    unknown0x132: ClassVar[UnsupportedMember] = UnsupportedMember("struct", 0x132)
    "Word. padding"
    AI: ClassVar[UnsupportedMember] = UnsupportedMember("struct", 0x134)  # Dword
    airStrength: ClassVar[UnsupportedMember] = UnsupportedMember("struct", 0x138)
    """Word"""
    groundStrength: ClassVar[UnsupportedMember] = UnsupportedMember("struct", 0x13A)
    """Word"""
    finderIndexLeft: ClassVar[UnsupportedMember] = UnsupportedMember("struct", 0x13C)
    """Dword"""
    finderIndexRight: ClassVar[UnsupportedMember] = UnsupportedMember(
        "struct", 0x140
    )  # Dword
    finderIndexTop: ClassVar[UnsupportedMember] = UnsupportedMember("struct", 0x144)
    """Dword"""
    finderIndexBottom: ClassVar[UnsupportedMember] = UnsupportedMember(
        "struct", 0x148
    )  # Dword
    repulseUnknown: ClassVar[ByteMember] = ByteMember("struct", 0x14C)
    "updated only when air unit is being pushed"
    repulseAngle: ClassVar[ByteMember] = ByteMember("struct", 0x14D)
    driftPos: ClassVar[WordMember] = WordMember("struct", 0x14E)
    "(mapsizex / 1.5 max)"
    driftX: ClassVar[ByteMember] = ByteMember("struct", 0x14E)
    driftY: ClassVar[ByteMember] = ByteMember("struct", 0x14F)

    @classproperty
    def range(self):
        return (EPD(0x59CCA8), EPD(0x59CCA8) + 84 * 1699, 84)

    def __init__(self, epd: int_or_var, *, ptr: int_or_var | None = None) -> None:
        """EPD Constructor of CUnit. Use CUnit.from_ptr(ptr) for ptr value"""
        _epd: int | c.EUDVariable
        self._ptr: int | c.EUDVariable | None

        if not isinstance(epd, CUnit):
            u, p = unProxy(epd), unProxy(ptr)
        else:
            u, p = epd._value, epd._ptr

        if isinstance(u, int):
            q, r = divmod(u - EPD(0x59CCA8), 84)  # check epd
            if r == 0 and 0 <= q < 1700:
                _epd, self._ptr = u, 0x59CCA8 + 336 * q
            else:
                raise EPError(_("Invalid input for CUnit: {}").format(epd))

            if p is not None and (not isinstance(p, int) or p != self._ptr):
                raise EPError(_("Invalid input for CUnit.ptr: {}").format(ptr))
        elif isinstance(u, c.EUDVariable):
            if p is None:
                self._ptr = None
                if EPDOffsetMap._cast:
                    _epd = u
                else:
                    _epd = c.EUDVariable()
                    _epd << u
            else:
                if not isinstance(p, c.EUDVariable):
                    raise EPError(_("Invalid input for CUnit.ptr: {}").format(ptr))
                if EPDOffsetMap._cast:
                    _epd, self._ptr = u, p
                else:
                    _epd, self._ptr = c.EUDCreateVariables(2)
                    c.SetVariables((_epd, self._ptr), (u, p))
        else:
            raise EPError(_("Invalid input for CUnit: {}").format(epd))

        EPDOffsetMap._cast = False
        super().__init__(_epd)

    @classmethod
    def from_ptr(cls, ptr: int_or_var) -> Self:
        epd: int | c.EUDVariable
        u = unProxy(ptr)
        # check ptr
        if isinstance(u, int):
            q, r = divmod(u - 0x59CCA8, 336)
            if r == 0 and 0 <= q < 1700:
                epd = EPD(u)
            else:
                raise EPError(_("Invalid input for CUnit: {}").format(ptr))
        elif isinstance(u, c.EUDVariable):
            epd = _epd_cache(u)
        else:
            raise EPError(_("Invalid input for CUnit: {}").format(epd))

        return cls(epd, ptr=u)

    @classmethod
    def from_read(cls, epd) -> Self:
        from ..memio import f_cunitepdread_epd

        _ptr, _epd = f_cunitepdread_epd(epd)
        return cls(_epd, ptr=_ptr)

    @classmethod
    def from_next(cls) -> Self:
        return cls.from_read(EPD(0x628438))

    @property
    def ptr(self) -> int | c.EUDVariable:
        if isinstance(self._value, int):
            return cast(int, self._ptr)  # FIXME
        if self._ptr is None:
            self._ptr = c.EUDVariable()
        return _ptr_cache(self._value, self._ptr)

    @staticmethod
    @c.EUDTypedFunc([None, None, TrgPlayer])
    def _cgive(unit, ptr, new_owner):
        from ..memio import f_cunitepdread_epd, f_dwwrite_epd, f_maskread_epd

        unit += 0x4C // 4  # playerID, orderID
        if cs.EUDIf()(c.MemoryXEPD(unit, c.Exactly, 0, 0xFF00)):
            c.EUDReturn()
        cs.EUDEndIf()

        prev_owner = f_maskread_epd(unit, 0xF)
        cs.DoActions(c.SetMemoryXEPD(unit, c.SetTo, new_owner, 0xFF))
        unit += 0x68 // 4 - 0x4C // 4  # prevPlayerUnit
        prev_ptr, prev_epd = f_cunitepdread_epd(unit)
        c.SetVariables(  # nextPlayerUnit
            [unit, prev_epd, prev_owner, new_owner],  # new_header
            [0x6C // 4 - 0x68 // 4, 0x6C // 4, EPD(0x6283F8), EPD(0x6283F8)],
            [c.Add, c.Add, c.Add, c.Add],
        )
        next_ptr, next_epd = f_cunitepdread_epd(unit)

        if cs.EUDIf()(prev_ptr >= 0x59CCA8):
            f_dwwrite_epd(prev_epd, next_ptr)
        if cs.EUDElse()():
            f_dwwrite_epd(EPD(0x6283F8) + prev_owner, next_ptr)
        cs.EUDEndIf()

        if cs.EUDIf()(next_ptr >= 0x59CCA8):
            next_epd += 0x68 // 4
            f_dwwrite_epd(next_epd, prev_ptr)
        cs.EUDEndIf()

        if cs.EUDIf()(c.MemoryEPD(new_owner, c.AtLeast, 0x59CCA8)):
            newprev_ptr, newprev_epd = f_cunitepdread_epd(new_owner)
            newprev_epd += 0x6C // 4
            newnext_ptr, newnext_epd = f_cunitepdread_epd(newprev_epd)
            f_dwwrite_epd(newprev_epd, ptr)
            f_dwwrite_epd(unit, newnext_ptr)
            unit -= 0x6C // 4 - 0x68 // 4  # prevPlayerUnit
            f_dwwrite_epd(unit, newprev_ptr)
            if cs.EUDIf()(newnext_ptr >= 0x59CCA8):
                newnext_epd += 0x68 // 4
                f_dwwrite_epd(newnext_epd, ptr)
            cs.EUDEndIf()
        if cs.EUDElse()():
            f_dwwrite_epd(new_owner, ptr)
            f_dwwrite_epd(unit, 0)
            unit -= 0x6C // 4 - 0x68 // 4  # prevPlayerUnit
            f_dwwrite_epd(unit, 0)
        cs.EUDEndIf()

    @staticmethod
    @c.EUDTypedFunc([None, None, TrgPlayer, None])
    def _cgive_subunit(unit, ptr, new_owner, ignore_subunit):
        from ..memio import f_cunitepdread_epd

        CUnit._cgive(unit, ptr, new_owner)
        unit += 0x70 // 4  # subunit
        if cs.EUDIf()([c.MemoryEPD(unit, c.AtLeast, 0x59CCA8), ignore_subunit == 0]):
            subunit_ptr, subunit = f_cunitepdread_epd(unit)
            CUnit._cgive(subunit, subunit_ptr, new_owner)
        cs.EUDEndIf()

    def cgive(self, new_owner, *, ignore_subunit=False) -> None:
        if ignore_subunit:
            CUnit._cgive_subunit(self, self.ptr, new_owner, ignore_subunit)
        else:
            CUnit._cgive(self, self.ptr, new_owner)

    @staticmethod
    @c.EUDTypedFunc([None, TrgPlayer])
    def _set_color(unit, color_player65536):
        from ..memio import f_maskwrite_epd, f_spriteepdread_epd

        color_epd = c.EUDVariable()
        check_sprite = c.MemoryEPD(0, c.Exactly, 0)
        c.VProc(
            unit,
            [
                unit.AddNumber(0x0C // 4),
                unit.SetDest(EPD(check_sprite) + 1),
            ],
        )
        if cs.EUDIfNot()(check_sprite):
            f_spriteepdread_epd(unit, ret=[EPD(check_sprite) + 2, color_epd])
        cs.EUDEndIf()
        if cs.EUDIfNot()(color_epd <= 2):
            f_maskwrite_epd(color_epd + 2, color_player65536, 0xFF0000)
        cs.EUDEndIf()

    def set_color(self, color_player) -> None:
        color_player = c.EncodePlayer(color_player)
        CUnit._set_color(self, color_player * 65536)

    def check_status_flag(self, value, mask=None) -> c.Condition:
        if mask is None:
            mask = value
        return c.MemoryXEPD(self._value + 0x0DC // 4, c.Exactly, value, mask)

    def set_status_flag(self, value, mask=None) -> None:
        from ..memio import f_maskwrite_epd

        if mask is None:
            mask = value
        f_maskwrite_epd(self._value + 0x0DC // 4, value, mask)

    def clear_status_flag(self, mask) -> None:
        self.set_status_flag(0, mask)

    def are_buildq_empty(self) -> list[c.Condition]:
        return [
            self.eqattr("buildQueue12", 0xE400E4),
            self.eqattr("buildQueue34", 0xE400E4),
            self.eqattr("buildQueue5", 0xE4),
        ]

    @staticmethod
    @_EUDPredefineReturn(2, 3)
    @c.EUDTypedFunc([None, TrgUnit])
    def _check_buildq(unit, unit_type):
        from ..memio import f_setcurpl2cpcache

        ret = CUnit._check_buildq._frets[0]

        check_bq0 = c.DeathsX(CurrentPlayer, c.Exactly, 0, 0, 0xFFFF)
        c.VProc(
            [unit, unit_type],
            [
                ret.SetNumber(0),
                c.SetMemory(0x6509B0, c.SetTo, 0x98 // 4),
                unit.QueueAddTo(EPD(0x6509B0)),
                unit_type.SetDest(EPD(check_bq0) + 2),
            ],
        )
        c.RawTrigger(conditions=check_bq0, actions=ret.SetNumber(1))
        if cs.EUDIfNot()(ret == 1):
            check_bq1 = c.DeathsX(CurrentPlayer, c.Exactly, 0, 0, 0xFFFF0000)
            check_bq2 = c.DeathsX(CurrentPlayer, c.Exactly, 0, 0, 0xFFFF)
            unit65536 = unit_type * 65536  # Does not change CurrentPlayer
            c.VProc(
                [unit65536, unit_type],
                [
                    unit65536.SetDest(EPD(check_bq1) + 2),
                    unit_type.SetDest(EPD(check_bq2) + 2),
                ],
            )
            c.RawTrigger(conditions=check_bq1, actions=ret.SetNumber(1))

            check_bq3 = c.DeathsX(CurrentPlayer, c.Exactly, 0, 0, 0xFFFF0000)
            check_bq4 = c.DeathsX(CurrentPlayer, c.Exactly, 0, 0, 0xFFFF)
            c.VProc(
                [unit65536, unit_type],
                [
                    c.SetMemory(0x6509B0, c.Add, 1),
                    unit65536.SetDest(EPD(check_bq3) + 2),
                    unit_type.SetDest(EPD(check_bq4) + 2),
                ],
            )
            c.RawTrigger(conditions=check_bq2, actions=ret.SetNumber(1))
            c.RawTrigger(conditions=check_bq3, actions=ret.SetNumber(1))
            c.RawTrigger(actions=c.SetMemory(0x6509B0, c.Add, 1))
            c.RawTrigger(conditions=check_bq4, actions=ret.SetNumber(1))
        cs.EUDEndIf()
        f_setcurpl2cpcache()
        # return False

    @staticmethod
    @_EUDPredefineReturn(1, 2)
    @c.EUDTypedFunc([None, TrgUnit, None])
    def _check_buildq_const(unit, unit_type, unit65536):
        from ..memio import f_setcurpl2cpcache

        ret = CUnit._check_buildq_const._frets[0]

        check_bq0 = c.DeathsX(CurrentPlayer, c.Exactly, 0, 0, 0xFFFF)
        check_bq1 = c.DeathsX(CurrentPlayer, c.Exactly, 0, 0, 0xFFFF0000)
        c.VProc(
            [unit, unit_type, unit65536],
            [
                ret.SetNumber(0),
                c.SetMemory(0x6509B0, c.SetTo, 0x98 // 4),
                unit.QueueAddTo(EPD(0x6509B0)),
                unit_type.SetDest(EPD(check_bq0) + 2),
                unit65536.SetDest(EPD(check_bq1) + 2),
            ],
        )
        c.RawTrigger(conditions=check_bq0, actions=ret.SetNumber(1))
        c.RawTrigger(conditions=check_bq1, actions=ret.SetNumber(1))

        check_bq2 = c.DeathsX(CurrentPlayer, c.Exactly, 0, 0, 0xFFFF)
        check_bq3 = c.DeathsX(CurrentPlayer, c.Exactly, 0, 0, 0xFFFF0000)
        c.VProc(
            [unit_type, unit65536],
            [
                c.SetMemory(0x6509B0, c.Add, 1),
                unit_type.SetDest(EPD(check_bq2) + 2),
                unit65536.SetDest(EPD(check_bq3) + 2),
            ],
        )
        c.RawTrigger(conditions=check_bq2, actions=ret.SetNumber(1))
        c.RawTrigger(conditions=check_bq3, actions=ret.SetNumber(1))

        check_bq4 = c.DeathsX(CurrentPlayer, c.Exactly, 0, 0, 0xFFFF)
        c.VProc(
            unit_type,
            [c.SetMemory(0x6509B0, c.Add, 1), unit_type.SetDest(EPD(check_bq4) + 2)],
        )
        c.RawTrigger(conditions=check_bq4, actions=ret.SetNumber(1))
        f_setcurpl2cpcache()
        # return False

    def check_buildq(self, unit_type) -> c.Condition:
        unit = c.EncodeUnit(unit_type)
        if isinstance(unit, int):
            return CUnit._check_buildq_const(self, unit, unit * 65536)
        else:
            return CUnit._check_buildq(self, unit)

    def reset_buildq(self, q1=0xE4) -> None:
        # See https://github.com/python/mypy/issues/14969
        self.buildQueue12 = 0xE40000 + q1  # type: ignore[misc]
        self.buildQueue34 = 0xE400E4  # type: ignore[misc]
        self.buildQueue5 = 0xE4  # type: ignore[misc]

    def die(self) -> None:
        # See https://github.com/python/mypy/issues/14969
        self.order = 0  # type: ignore[misc]

    def remove_collision(self) -> None:
        self.set_status_flag(0x00A00000)

    def set_collision(self) -> None:
        self.clear_status_flag(0x00A00000)

    def set_invincible(self) -> None:
        self.set_status_flag(0x04000000)

    def clear_invincible(self) -> None:
        self.clear_status_flag(0x04000000)

    def set_gathering(self) -> None:
        self.set_status_flag(0x00800000)

    def clear_gathering(self) -> None:
        self.clear_status_flag(0x00800000)

    def set_speed_upgrade(self) -> None:
        self.set_status_flag(0x10000000)

    def clear_speed_upgrade(self) -> None:
        self.clear_status_flag(0x10000000)

    def set_hallucination(self) -> None:
        self.set_status_flag(0x40000000)

    def clear_hallucination(self) -> None:
        self.clear_status_flag(0x40000000)

    def power(self) -> None:
        self.clear_status_flag(0x00000008)

    def unpower(self) -> None:
        self.set_status_flag(0x00000008)

    def set_air(self) -> None:
        self.set_status_flag(0x00000004)

    def set_ground(self) -> None:
        self.clear_status_flag(0x00000004)

    def clear_noclip(self) -> None:
        self.set_status_flag(0x00100000)

    def set_noclip(self) -> None:
        self.clear_status_flag(0x00100000)

    def is_dying(self) -> tuple[c.Condition, c.Condition]:
        from ..eudlib.utilf.unlimiterflag import IsUnlimiterOn

        if IsUnlimiterOn():
            raise EPError(_("Can't detect unit dying with [unlimiter]"))
        return (  # return (self.order == 0, self.sprite >= 1)
            c.MemoryXEPD(self._value + 0x4D // 4, c.Exactly, 0, 0xFF00),
            c.MemoryEPD(self._value + 0x0C // 4, c.AtLeast, 1),
        )

    def is_completed(self) -> c.Condition:
        return self.check_status_flag(0x00000001)

    def is_hallucination(self) -> c.Condition:
        return self.check_status_flag(0x40000000)

    def is_air(self):
        return self.check_status_flag(0x00000004)

    def is_in_building(self) -> c.Condition:
        return self.check_status_flag(0x00000020)

    def is_in_transport(self) -> c.Condition:
        return self.check_status_flag(0x00000040)

    def is_burrowed(self) -> c.Condition:
        return self.check_status_flag(0x00000010)

    def setloc(self, location) -> None:
        from ..eudlib.locf.locf import f_setloc_epd

        f_setloc_epd(location, self._value + 0x28 // 4)

    def remove(self) -> None:
        # See https://github.com/python/mypy/issues/14969
        self.userActionFlags = 4  # type: ignore[misc]
        self.order = 0  # type: ignore[misc]


EPDCUnitMap = CUnit



# Copyright 2023 by Armoha.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from abc import ABCMeta, abstractmethod
from typing import Literal

from ...core import (
    EncodeButtonSet,
    EncodeIcon,
    EncodeIscript,
    EncodePortrait,
    EncodeString,
    EncodeTBL,
    EUDVariable,
)
from ...core.rawtrigger.strdict import DefRankDict, DefSfxDataDict
from ...core.rawtrigger.strenc import _EncodeAny
from ...utils import ep_assert, unProxy


class BaseKind(metaclass=ABCMeta):
    __slots__ = ()

    @classmethod
    def cast(cls, other):
        return other

    @classmethod
    @abstractmethod
    def size(cls) -> Literal[1, 2, 4]:
        ...

    @classmethod
    @abstractmethod
    def read_epd(cls, epd, subp):
        ...

    @classmethod
    @abstractmethod
    def write_epd(cls, epd, subp, value) -> None:
        ...

    @classmethod
    @abstractmethod
    def add_epd(cls, epd, subp, value) -> None:
        ...

    @classmethod
    @abstractmethod
    def subtract_epd(cls, epd, subp, value) -> None:
        ...


class ByteKind(BaseKind):
    __slots__ = ()

    @classmethod
    def size(cls) -> Literal[1]:
        return 1

    @classmethod
    def read_epd(cls, epd, subp) -> EUDVariable:
        from ...memio import f_bread_epd

        return f_bread_epd(epd, subp)

    @classmethod
    def write_epd(cls, epd, subp, value) -> None:
        from ...memio import f_bwrite_epd

        f_bwrite_epd(epd, subp, value)

    @classmethod
    def add_epd(cls, epd, subp, value) -> None:
        from ...memio import f_badd_epd

        f_badd_epd(epd, subp, value)

    @classmethod
    def subtract_epd(cls, epd, subp, value) -> None:
        from ...memio import f_bsubtract_epd

        f_bsubtract_epd(epd, subp, value)


class Bit0Kind(ByteKind):
    __slots__ = ()

    @classmethod
    def read_epd(cls, epd, subp) -> EUDVariable:
        from ...memio.bwepdio import _bitread_epd

        return _bitread_epd(0x01)(epd, subp)

    @classmethod
    def write_epd(cls, epd, subp, value) -> None:
        from ...memio.bwepdio import _bitwrite_epd

        _bitwrite_epd(epd, subp, 0x01, value)

    @classmethod
    def add_epd(cls, epd, subp, value) -> None:
        raise NotImplementedError

    @classmethod
    def subtract_epd(cls, epd, subp, value) -> None:
        raise NotImplementedError


class Bit1Kind(ByteKind):
    __slots__ = ()

    @classmethod
    def read_epd(cls, epd, subp) -> EUDVariable:
        from ...memio.bwepdio import _bitread_epd

        return _bitread_epd(0x02)(epd, subp)

    @classmethod
    def write_epd(cls, epd, subp, value) -> None:
        from ...memio.bwepdio import _bitwrite_epd

        _bitwrite_epd(epd, subp, 0x02, value)

    @classmethod
    def add_epd(cls, epd, subp, value) -> None:
        raise NotImplementedError

    @classmethod
    def subtract_epd(cls, epd, subp, value) -> None:
        raise NotImplementedError


class PlayerKind(ByteKind):
    __slots__ = ()

    @classmethod
    def cast(cls, other):
        from ..player import TrgPlayer

        return TrgPlayer.cast(other)

    @classmethod
    def read_epd(cls, epd, subp) -> EUDVariable:
        from ...memio.specialized import _playerread_epd

        return _playerread_epd()[subp](epd)


class WeaponKind(ByteKind):
    __slots__ = ()

    @classmethod
    def cast(cls, other):
        from ..weapon import Weapon

        return Weapon.cast(other)


class UnitOrderKind(ByteKind):
    __slots__ = ()

    @classmethod
    def cast(cls, other):
        from ..unitorder import UnitOrder

        return UnitOrder.cast(other)


class UpgradeKind(ByteKind):
    __slots__ = ()

    @classmethod
    def cast(cls, other):
        from ..upgrade import Upgrade

        return Upgrade.cast(other)


class TechKind(ByteKind):
    __slots__ = ()

    @classmethod
    def cast(cls, other):
        from ..tech import Tech

        return Tech.cast(other)


class UnitSizeKind(ByteKind):
    __slots__ = ()

    @classmethod
    def cast(cls, other):
        return _EncodeAny(
            "UnitSize",
            lambda s: {}[s],
            {
                "Independent": 0,
                "Small": 1,
                "Medium": 2,
                "Large": 3,
            },
            other,
        )


class RightClickActionKind(ByteKind):
    __slots__ = ()

    @classmethod
    def cast(cls, other):
        return _EncodeAny(
            "RightClickAction",
            lambda s: {}[s],
            {
                "NoCommand_AutoAttack": 0,
                "NormalMove_NormalAttack": 1,
                "NormalMove_NoAttack": 2,
                "NoMove_NormalAttack": 3,
                "Harvest": 4,
                "HarvestAndRepair": 5,
                "Nothing": 6,
            },
            other,
        )


class MovementControlKind(ByteKind):
    __slots__ = ()

    @classmethod
    def cast(cls, other):
        return _EncodeAny(
            "MovementControl",
            lambda s: {}[s],
            {
                "FlingyDat": 0,
                "PartiallyMobile_Weapon": 1,
                "IscriptBin": 2,
            },
            other,
        )


class DamageTypeKind(ByteKind):
    __slots__ = ()

    @classmethod
    def cast(cls, other):
        return _EncodeAny(
            "DamageType",
            lambda s: {}[s],
            {
                "Independent": 0,
                "Explosive": 1,
                "Concussive": 2,
                "Normal": 3,
                "IgnoreArmor": 4,
            },
            other,
        )


class WeaponBehaviorKind(ByteKind):
    __slots__ = ()

    @classmethod
    def cast(cls, other):
        return _EncodeAny(
            "WeaponBehavior",
            lambda s: {}[s],
            {
                "Fly_DoNotFollowTarget": 0,
                "Fly_FollowTarget": 1,
                "AppearOnTargetUnit": 2,
                "PersistOnTargetSite": 3,  # Psionic Storm
                "AppearOnTargetSite": 4,
                "AppearOnAttacker": 5,
                "AttackAndSelfDestruct": 6,
                "Bounce": 7,  # Mutalisk Glave Wurms
                "AttackNearbyArea": 8,  # Valkyrie Halo Rockets
                "GoToMaxRange": 9,  # Lurker Subterranean Spines
            },
            other,
        )


class ExplosionTypeKind(ByteKind):
    __slots__ = ()

    @classmethod
    def cast(cls, other):
        return _EncodeAny(
            "ExplosionType",
            lambda s: {}[s],
            {
                "None": 0,
                "NormalHit": 1,
                "SplashRadial": 2,
                "SplashEnemy": 3,
                "Lockdown": 4,
                "NuclearMissile": 5,
                "Parasite": 6,
                "Broodlings": 7,
                "EmpShockwave": 8,
                "Irradiate": 9,
                "Ensnare": 10,
                "Plague": 11,
                "StasisField": 12,
                "DarkSwarm": 13,
                "Consume": 14,
                "YamatoGun": 15,
                "Restoration": 16,
                "DisruptionWeb": 17,
                "CorrosiveAcid": 18,
                "MindControl": 19,
                "Feedback": 20,
                "OpticalFlare": 21,
                "Maelstrom": 22,
                "Unknown_Crash": 23,
                "SplashAir": 24,
            },
            other,
        )


class DrawingFunctionKind(ByteKind):
    __slots__ = ()

    @classmethod
    def cast(cls, other):
        return _EncodeAny(
            "DrawingFunction",
            lambda s: {}[s],
            {
                "Normal": 0,
                "NormalNoHallucination": 1,
                "NonVisionCloaking": 2,
                "NonVisionCloaked": 3,
                "NonVisionDecloaking": 4,
                "VisionCloaking": 5,
                "VisionCloaked": 6,
                "VisionDecloaking": 7,
                "EMPShockwave": 8,
                "UseRemapping": 9,
                "Shadow": 10,
                "HpBar": 11,  # crash
                "WarpTexture": 12,
                "SelectionCircle": 13,
                "PlayerColorOverride": 14,  # used for flags
                "HideGFX_ShowSizeRect": 15,
                "Hallucination": 16,
                "WarpFlash": 17,
            },
            other,
        )


class AnimationKind(ByteKind):
    __slots__ = ()

    @classmethod
    def cast(cls, other):
        return _EncodeAny(
            "IscriptAnimation",
            lambda s: {}[s],
            {
                "Init": 0,
                "Death": 1,
                "GndAttkInit": 2,
                "AirAttkInit": 3,
                "Unused1": 4,
                "GndAttkRpt": 5,
                "AirAttkRpt": 6,
                "CastSpell": 7,
                "GndAttkToIdle": 8,
                "AirAttkToIdle": 9,
                "Unused2": 10,
                "Walking": 11,
                "WalkingToIdle": 12,
                "SpecialState1": 13,
                "SpecialState2": 14,
                "AlmostBuilt": 15,
                "Built": 16,
                "Landing": 17,
                "LiftOff": 18,
                "IsWorking": 19,
                "WorkingToIdle": 20,
                "WarpIn": 21,
                "Unused3": 22,
                "StarEditInit": 23,
                "Disable": 24,
                "Burrow": 25,
                "UnBurrow": 26,
                "Enable": 27,
                "NoAnimation": 28,
            },
            other,
        )


class RaceResearchKind(ByteKind):
    __slots__ = ()

    @classmethod
    def cast(cls, other):
        return _EncodeAny(
            "race",
            lambda s: {}[s],
            {
                "Zerg": 0,
                "Terran": 1,
                "Protoss": 2,
                "All": 3,
            },
            other,
        )


class WorkerCarryTypeKind(ByteKind):
    __slots__ = ()

    @classmethod
    def cast(cls, other):
        return _EncodeAny(
            "CarryType",
            lambda s: {}[s],
            {
                "None": 0,
                "Gas": 1,
                "Ore": 2,
                "GasOrOre": 3,
                "PowerUp": 4,
            },
            other,
        )


class RankKind(ByteKind):
    __slots__ = ()

    @classmethod
    def cast(cls, other):
        return _EncodeAny(
            "rank",
            lambda s: {}[s],
            DefRankDict,
            other,
        )


class FlingyKind(ByteKind):
    """size of Flingy is byte in units.dat / word in CUnit"""

    __slots__ = ()

    @classmethod
    def cast(cls, other):
        from ..flingy import Flingy

        return Flingy.cast(other)


class WordKind(BaseKind):
    __slots__ = ()

    @classmethod
    def size(cls) -> Literal[2]:
        return 2

    @classmethod
    def read_epd(cls, epd, subp) -> EUDVariable:
        from ...memio import f_wread_epd

        return f_wread_epd(epd, subp)

    @classmethod
    def write_epd(cls, epd, subp, value) -> None:
        from ...memio import f_wwrite_epd

        f_wwrite_epd(epd, subp, value)

    @classmethod
    def add_epd(cls, epd, subp, value) -> None:
        from ...memio import f_wadd_epd

        f_wadd_epd(epd, subp, value)

    @classmethod
    def subtract_epd(cls, epd, subp, value) -> None:
        from ...memio import f_wsubtract_epd

        f_wsubtract_epd(epd, subp, value)


class PositionXKind(WordKind):
    __slots__ = ()

    @classmethod
    def read_epd(cls, epd, subp) -> EUDVariable:
        from ...memio.specialized import _mapxread_epd

        return _mapxread_epd()[subp // 2](epd)


class PositionYKind(WordKind):
    __slots__ = ()

    @classmethod
    def read_epd(cls, epd, subp) -> EUDVariable:
        from ...memio.specialized import _mapyread_epd

        return _mapyread_epd()[subp // 2](epd)


class UnitKind(WordKind):
    __slots__ = ()

    @classmethod
    def cast(cls, other):
        from ..unit import TrgUnit

        return TrgUnit.cast(other)

    @classmethod
    def read_epd(cls, epd, subp) -> EUDVariable:
        from ...memio import f_bread_epd

        return f_bread_epd(epd, subp)


class SpriteKind(WordKind):
    __slots__ = ()

    @classmethod
    def cast(cls, other):
        from ..sprite import Sprite

        return Sprite.cast(other)


class ImageKind(WordKind):
    __slots__ = ()

    @classmethod
    def cast(cls, other):
        from ..image import Image

        return Image.cast(other)


class StatTextKind(WordKind):
    __slots__ = ()

    @classmethod
    def cast(cls, other):
        return EncodeTBL(other)


class SfxDataKind(WordKind):
    __slots__ = ()

    @classmethod
    def cast(cls, other):
        other = unProxy(other)
        if isinstance(other, str):
            other = other.replace("\\", "/").lower()
        return _EncodeAny(
            "sfxdata.dat",
            lambda s: {}[s],
            DefSfxDataDict,
            other,
        )


class PortraitKind(WordKind):
    __slots__ = ()

    @classmethod
    def cast(cls, other):
        return EncodePortrait(other)


class IconKind(WordKind):
    __slots__ = ()

    @classmethod
    def cast(cls, other):
        return EncodeIcon(other)


class MapStringKind(WordKind):
    __slots__ = ()

    @classmethod
    def cast(cls, other):
        return EncodeString(other)


class ButtonSetKind(WordKind):
    __slots__ = ()

    @classmethod
    def cast(cls, other):
        return EncodeButtonSet(other)


class DwordKind(BaseKind):
    __slots__ = ()

    @classmethod
    def size(cls) -> Literal[4]:
        return 4

    @classmethod
    def read_epd(cls, epd, subp) -> EUDVariable:
        from ...memio import f_dwread_epd

        ep_assert(isinstance(subp, int) and subp == 0, "malaligned dword member")
        return f_dwread_epd(epd)

    @classmethod
    def write_epd(cls, epd, subp, value) -> None:
        from ...memio import f_dwwrite_epd

        ep_assert(isinstance(subp, int) and subp == 0, "malaligned dword member")
        f_dwwrite_epd(epd, value)

    @classmethod
    def add_epd(cls, epd, subp, value) -> None:
        from ...memio import f_dwadd_epd

        ep_assert(isinstance(subp, int) and subp == 0, "malaligned dword member")
        f_dwadd_epd(epd, value)

    @classmethod
    def subtract_epd(cls, epd, subp, value) -> None:
        from ...memio import f_dwsubtract_epd

        ep_assert(isinstance(subp, int) and subp == 0, "malaligned dword member")
        f_dwsubtract_epd(epd, value)


class PositionKind(DwordKind):
    __slots__ = ()

    @classmethod
    def read_epd(cls, epd, subp) -> EUDVariable:
        from ...memio import f_maskread_epd
        from ...memio.specialized import _map_xy_mask

        ep_assert(isinstance(subp, int) and subp == 0)
        return f_maskread_epd(epd, (lambda x, y: x + 65536 * y)(*_map_xy_mask()))


class CUnitKind(DwordKind):
    __slots__ = ()

    @classmethod
    def read_epd(cls, epd, subp):
        from ..cunit import CUnit

        ep_assert(isinstance(subp, int) and subp == 0, "malaligned CUnit member")
        return CUnit.from_read(epd)


class CSpriteKind(DwordKind):
    __slots__ = ()

    @classmethod
    def read_epd(cls, epd, subp):
        from ..csprite import CSprite

        ep_assert(isinstance(subp, int) and subp == 0, "malaligned CSprite member")
        return CSprite.from_read(epd)


class IscriptKind(DwordKind):
    __slots__ = ()

    @classmethod
    def cast(cls, other):
        return EncodeIscript(other)


class SelfKind(BaseKind):
    __slots__ = ()



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from ...localize import _
from ...utils import EPD, ep_assert, ep_warn
from ..mapdata import UnitProperty
from .action import Action
from .constenc import (
    Add,
    EncodeAllyStatus,
    EncodeCount,
    EncodeModifier,
    EncodeOrder,
    EncodePlayer,
    EncodeProperty,
    EncodePropState,
    EncodeResource,
    EncodeScore,
    EncodeSwitchAction,
)
from .consttype import Byte, Dword
from .strenc import (
    EncodeAIScript,
    EncodeLocation,
    EncodeString,
    EncodeSwitch,
    EncodeUnit,
)
from .typehint import (
    AIScriptAtLocation,
    AIScriptWithoutLocation,
    AllyStatus,
    Count,
    Location,
    Modifier,
    Player,
    PropState,
    Resource,
    String,
    SwitchAction,
    Unit,
    _Order,
    _Score,
    _Switch,
)


def Victory() -> Action:  # noqa: N802
    """End scenario in victory for current player.

    The game ends in victory for the trigger's owner.
    Any players who are not executing a victory action are defeated.
    """
    return Action(0, 0, 0, 0, 0, 0, 0, 1, 0, 4)


def Defeat() -> Action:  # noqa: N802
    """End scenario in defeat for current player.

    This will end the scenario in defeat for the affected players.
    Any other players in the game will continue.
    """
    return Action(0, 0, 0, 0, 0, 0, 0, 2, 0, 4)


def PreserveTrigger() -> Action:  # noqa: N802
    """Preserve Trigger.

    Normally, a trigger will only run once for each owner.
    Triggers automatically disable themselves once they run through all of
    their actions, unless the Preserve Trigger action is present.
    If you want a trigger to remain in effect throughout the scenario,
    add the Preserve Trigger action to its action list.
    """
    return Action(0, 0, 0, 0, 0, 0, 0, 3, 0, 4)


def Wait(time: Dword) -> Action:  # noqa: N802
    """Wait for duration milliseconds.

    The wait action is used to delay other actions for the specified number of
    milliseconds. Because it is a blocking action, no other actions in the same
    trigger and no other blocking actions in other triggers will activate
    until it is done.
    """
    ep_warn(_("Don't use Wait action UNLESS YOU KNOW WHAT YOU'RE DOING!"))
    return Action(0, 0, 0, time, 0, 0, 0, 4, 0, 4)


def PauseGame() -> Action:  # noqa: N802
    """Pause the game.

    This action will put the game in a pause state.
    If a matching Unpause Game is not found, the program automatically unpauses
    the game when the current trigger is finished. Note that pause game has no
    effect in multiplayer scenarios or against computer controlled players.
    """
    return Action(0, 0, 0, 0, 0, 0, 0, 5, 0, 4)


def UnpauseGame() -> Action:  # noqa: N802
    """Unpause the game.

    This action resumes the game from a paused session.
    Note that this has no effect in multiplayer maps and will not effect any
    computer opponents in single player maps.
    """
    return Action(0, 0, 0, 0, 0, 0, 0, 6, 0, 4)


def Transmission(  # noqa: N802
    unit: Unit,
    where: Location,
    sound_name: String,
    time_modifier: Modifier,
    time: Dword,
    text: String,
    AlwaysDisplay: Byte = 4,  # noqa: N803
) -> Action:
    """Send transmission to current player from unit at location.

    A transmission is a combination of several different actions.
    First, you need to specify which unit at a location you want to send the
    transmission. This unit's portrait will be displayed for the duration of
    the transmission.
    You then need to select a WAV file to play, how long to animate the unit
    portrait, and what text message to display. The player receiving
    the transmission will receive a minimap ping when the transmission starts,
    and can press the space bar to center their screen on the unit sending the
    transmission.

    Note that this action has no effect on computer players,
    and it will prevent any other action (in the same trigger) from resuming
    until it has finished.
    """
    from ...epscript.epsimp import IsSCDBMap

    ep_assert(
        isinstance(AlwaysDisplay, int),
        _("AlwaysDisplay argument must be int, not '{}'").format(AlwaysDisplay),
    )
    if not IsSCDBMap():
        ep_warn(_("Don't use Wait action UNLESS YOU KNOW WHAT YOU'RE DOING!"))
    unit = EncodeUnit(unit)
    where = EncodeLocation(where)
    sound_name = EncodeString(sound_name)
    modifier = EncodeModifier(time_modifier)
    text = EncodeString(text)
    return Action(where, text, sound_name, time, 0, 0, unit, 7, modifier, 4)


def PlayWAV(sound_name: String) -> Action:  # noqa: N802
    """Play WAV file.

    This will play a WAV file for the trigger's owner.
    """
    sound_name = EncodeString(sound_name)
    return Action(0, 0, sound_name, 0, 0, 0, 0, 8, 0, 4)


def DisplayText(text: String, AlwaysDisplay: Byte = 4) -> Action:  # noqa: N802, N803
    """Display for current player: text.

    Displays a specific text message to each owner of the condition.
    """
    ep_assert(
        isinstance(AlwaysDisplay, int),
        _("AlwaysDisplay argument must be int, not '{}'").format(AlwaysDisplay),
    )
    text = EncodeString(text)
    return Action(0, text, 0, 0, 0, 0, 0, 9, 0, 4)


def CenterView(where: Location) -> Action:  # noqa: N802
    """Center view for current player at location.

    This action creates the specified number of units at the specified Location
    If a unit is created 'Anywhere', it will appear in the center of the map.
    This action will not function while the game is paused.

    Keep in mind that when the conditions are successfully met,
    the unit(s) will be created for each player that owns the trigger.
    For example, if All Players own a trigger that creates a Terran Marine for
    Player 1, and the conditions of the trigger are true for four of the
    players, Player 1 will get 4 Marines.
    """
    where = EncodeLocation(where)
    return Action(where, 0, 0, 0, 0, 0, 0, 10, 0, 4)


def CreateUnitWithProperties(  # noqa: N802
    count: Byte,
    unit: Unit,
    where: Location,
    player: Player,
    properties: UnitProperty | bytes,
) -> Action:
    """Create quantity unit at location for player. Apply properties

    This action works just like Create Unit, except that you can customize
    the properties of the newly created unit(s).
    """
    if isinstance(Count, int):
        ep_assert(0 <= Count <= 255)
    unit = EncodeUnit(unit)
    where = EncodeLocation(where)
    player = EncodePlayer(player)
    _property = EncodeProperty(properties)
    return Action(where, 0, 0, 0, player, _property, unit, 11, count, 28)


def SetMissionObjectives(text: String) -> Action:  # noqa: N802
    """Set Mission Objectives to: text.

    Changes the mission objectives text to something other than what was
    defined at the outset of the level. While this doesn't actually change the
    victory or defeat conditions for the scenario, it can be used to notify
    the players of changes to the scenario's objectives.
    """
    text = EncodeString(text)
    return Action(0, text, 0, 0, 0, 0, 0, 12, 0, 4)


def SetSwitch(switch: _Switch, state: SwitchAction) -> Action:  # noqa: N802
    """Set switch.

    The set switch action can be used to:
    - Set a switch to its set position.
    - Clear a switch to its cleared position.
    - Toggle a switch: if a switch is cleared, it becomes set;
                       if it is set, it becomes cleared.
    - Randomly choose between the set or cleared position.
    """
    switch = EncodeSwitch(switch)
    state = EncodeSwitchAction(state)
    return Action(0, 0, 0, 0, 0, switch, 0, 13, state, 4)


def SetCountdownTimer(time_modifier: Modifier, time: Dword) -> Action:  # noqa: N802
    """Modify Countdown Timer: Set duration seconds.

    This allows you to set a countdown timer, in game seconds, which will
    appear at the top of the game screen and count down automatically.
    There is one countdown timer shared by all players. Any time the countdown
    timer is not equal to zero, it is displayed to all players.

    You can also use this action to add or subtract time from the countdown
    timer.
    """
    modifier = EncodeModifier(time_modifier)
    return Action(0, 0, 0, time, 0, 0, 0, 14, modifier, 4)


def RunAIScript(script: AIScriptWithoutLocation) -> Action:  # noqa: N802
    """Execute AI script script.

    This instructs the specified computer-controlled players to use a certain
    AI script. The AI script determines the overall aggressiveness and
    effectiveness of the computer player, and by changing the AI script during
    the scenario, you can effectively handicap the scenario.
    """
    script = EncodeAIScript(script)
    return Action(0, 0, 0, 0, 0, script, 0, 15, 0, 4)


def RunAIScriptAt(script: AIScriptAtLocation, where: Location) -> Action:  # noqa: N802
    """Execute AI script script at location.

    Identical to [Run AI Script](#link_action_runaiscript) but specifies
    a location to run the script at.
    Certain scripts are designed specifically to target a Location.
    """
    script = EncodeAIScript(script)
    where = EncodeLocation(where)
    return Action(where, 0, 0, 0, 0, script, 0, 16, 0, 4)


def LeaderBoardControl(unit: Unit, label: String) -> Action:  # noqa: N802
    """Show Leader Board for most control of unit. Display label: label

    This will display the Leader Board to all players based on who controls
    the most of a particular unit in the scenario.
    """
    unit = EncodeUnit(unit)
    label = EncodeString(label)
    return Action(0, label, 0, 0, 0, 0, unit, 17, 0, 20)


def LeaderBoardControlAt(  # noqa: N802
    unit: Unit, location: Location, label: String
) -> Action:
    """Show Leader Board for most control of units at location.

    Display label: label
    """
    unit = EncodeUnit(unit)
    location = EncodeLocation(location)
    label = EncodeString(label)
    return Action(location, label, 0, 0, 0, 0, unit, 18, 0, 20)


def LeaderBoardResources(resource_type: Resource, label: String) -> Action:  # noqa: N802
    """Show Leader Board for accumulation of most resource.

    Display label: label
    This will display the Leader Board to all players based on who has
    the most resources.
    """
    resource_type = EncodeResource(resource_type)
    label = EncodeString(label)
    return Action(0, label, 0, 0, 0, 0, resource_type, 19, 0, 4)


def LeaderBoardKills(unit: Unit, label: String) -> Action:  # noqa: N802
    """Show Leader Board for most kills of unit. Display label: label

    This will display the Leader Board to all players based on who has the most
    kills in the scenario.
    """
    unit = EncodeUnit(unit)
    label = EncodeString(label)
    return Action(0, label, 0, 0, 0, 0, unit, 20, 0, 20)


def LeaderBoardScore(ScoreType: _Score, label: String) -> Action:  # noqa: N803, N802
    """Show Leader Board for most points. Display label: label

    This will display the Leader Board to all players based on who has
    the most points.
    """
    score_type = EncodeScore(ScoreType)
    label = EncodeString(label)
    return Action(0, label, 0, 0, 0, 0, score_type, 21, 0, 4)


def KillUnit(unit: Unit, player: Player) -> Action:  # noqa: N802
    """Kill all units for player.

    This action kills all units of a particular type for the player specified.
    This action has no effect while the game is paused.
    """
    unit = EncodeUnit(unit)
    player = EncodePlayer(player)
    return Action(0, 0, 0, 0, player, 0, unit, 22, 0, 20)


def KillUnitAt(  # noqa: N802
    count: Count, unit: Unit, where: Location, for_player: Player
) -> Action:
    """Kill quantity units for player at location.

    Similar to the 'Kill Unit' action, the 'Kill Unit at Location' action gives
    you the ability to kill a specified number of units of a particular type
    belonging to a certain player at the specified Location.
    This action will not function while the game is paused.
    """
    count = EncodeCount(count)
    unit = EncodeUnit(unit)
    where = EncodeLocation(where)
    for_player = EncodePlayer(for_player)
    return Action(where, 0, 0, 0, for_player, 0, unit, 23, count, 20)


def RemoveUnit(unit: Unit, player: Player) -> Action:  # noqa: N802
    """Remove all units for player.

    Remove Unit works just like Kill Unit, except that the affected units will
    simply disappear without actually dying.
    This action has no effect while the game is paused.
    """
    unit = EncodeUnit(unit)
    player = EncodePlayer(player)
    return Action(0, 0, 0, 0, player, 0, unit, 24, 0, 20)


def RemoveUnitAt(  # noqa: N802
    count: Count, unit: Unit, where: Location, for_player: Player
) -> Action:
    """Remove all units for player.

    This action works just like Remove Unit.
    In addition, you may specify a location and a quantity of units that the
    action will affect. It has no effect on a paused game.
    """
    count = EncodeCount(count)
    unit = EncodeUnit(unit)
    where = EncodeLocation(where)
    for_player = EncodePlayer(for_player)
    return Action(where, 0, 0, 0, for_player, 0, unit, 25, count, 20)


def SetResources(  # noqa: N802
    player: Player, modifier: Modifier, amount: Dword, resource_type: Resource
) -> Action:
    """Modify resources for player: Set quantity resource.

    The set resources action allows you to increase, decrease, or set
    the amount of resources that a player has.
    """
    player = EncodePlayer(player)
    modifier = EncodeModifier(modifier)
    resource_type = EncodeResource(resource_type)
    return Action(0, 0, 0, 0, player, amount, resource_type, 26, modifier, 4)


def SetScore(  # noqa: N802
    player: Player,
    modifier: Modifier,
    amount: Dword,
    ScoreType: _Score,  # noqa: N803
) -> Action:
    """Modify score for player: Set quantity points.

    The set score action lets you the increase, decrease, or set the number of
    points that a player currently has.
    """
    player = EncodePlayer(player)
    modifier = EncodeModifier(modifier)
    score_type = EncodeScore(ScoreType)
    return Action(0, 0, 0, 0, player, amount, score_type, 27, modifier, 4)


def MinimapPing(where: Location) -> Action:  # noqa: N802
    """Show minimap ping for current player at location.

    This sends out a 'ping' on the mini map at the specified location.
    This can be used to draw attention to a particular spot or to track a
    moving location. Note that pressing the spacebar in the game after
    receiving the ping will not center your screen on the ping Location.
    Only transmissions allow you to jump to a different location with the
    spacebar.
    """
    where = EncodeLocation(where)
    return Action(where, 0, 0, 0, 0, 0, 0, 28, 0, 4)


def TalkingPortrait(unit: Unit, time: Dword) -> Action:  # noqa: N802
    """Show unit talking to current player for duration milliseconds.

    This will show the unit picture of your choice in the unit window in the
    game screen for the specified amount of time.
    """
    unit = EncodeUnit(unit)
    return Action(0, 0, 0, time, 0, 0, unit, 29, 0, 20)


def MuteUnitSpeech() -> Action:  # noqa: N802
    """Mute all non-trigger unit sounds for current player.

    This action will mute unit speech and set to half-volume all sound effects
    that the game normally produces, including music and combat sounds.
    This is particularly useful when you are playing a Transmission Action
    or any time you want to make sure a triggered sound is heard clearly.
    """
    return Action(0, 0, 0, 0, 0, 0, 0, 30, 0, 4)


def UnMuteUnitSpeech() -> Action:  # noqa: N802
    """Unmute all non-trigger unit sounds for current player.

    This action sets the sound effects for the game back to their original
    state.
    """
    return Action(0, 0, 0, 0, 0, 0, 0, 31, 0, 4)


def LeaderBoardComputerPlayers(state: PropState) -> Action:  # noqa: N802
    """Set use of computer players in leaderboard calculations.

    This action allows you to specify whether neutral, rescue and computer
    controlled players will be included in the leader board calculations.
    By default, all computer players are included in the tally.
    """
    state = EncodePropState(state)
    return Action(0, 0, 0, 0, 0, 0, 0, 32, state, 4)


def LeaderBoardGoalControl(goal: Dword, unit: Unit, label: String) -> Action:  # noqa: N802
    """Show Leader Board for player closest to control of number of unit.

    Display label: label
    This will display the Leader Board to all players based on the amount of
    units controlled on the map that are required to achieve a goal.
    In this type of leader board, the lower the number the better.
    """
    unit = EncodeUnit(unit)
    label = EncodeString(label)
    return Action(0, label, 0, 0, 0, goal, unit, 33, 0, 20)


def LeaderBoardGoalControlAt(  # noqa: N802
    goal: Dword, unit: Unit, location: Location, label: String
) -> Action:
    """Show Leader Board for player closest to control of number of units at
    location.

    Display label: label
    This will display the Leader Board to all players based on the amount of
    units controlled at a certain Location that are required to achieve a goal.
    In this type of leader board, the lower the number the better.
    """
    unit = EncodeUnit(unit)
    location = EncodeLocation(location)
    label = EncodeString(label)
    return Action(location, label, 0, 0, 0, goal, unit, 34, 0, 20)


def LeaderBoardGoalResources(  # noqa: N802
    goal: Dword, resource_type: Resource, label: String
) -> Action:
    """Show Leader Board for player closest to accumulation of number resource.

    Display label: label
    This will display the Leader Board to all players based on who have
    the most resources required to achieve a goal.
    In this type of leader board, the lower the number the better.
    """
    resource_type = EncodeResource(resource_type)
    label = EncodeString(label)
    return Action(0, label, 0, 0, 0, goal, resource_type, 35, 0, 4)


def LeaderBoardGoalKills(goal: Dword, unit: Unit, label: String) -> Action:  # noqa: N802
    """Show Leader Board for player closest to number kills of unit.

    Display label: label
    This will display the Leader Board to all players based on who have
    the most kills required to achieve a goal.
    In this type of leader board, the lower the number the better.
    """
    unit = EncodeUnit(unit)
    label = EncodeString(label)
    return Action(0, label, 0, 0, 0, goal, unit, 36, 0, 20)


def LeaderBoardGoalScore(  # noqa: N802
    goal: Dword,
    ScoreType: _Score,  # noqa: N803
    label: String,
) -> Action:
    """Show Leader Board for player closest to number points.

    Display label: label
    This will display the Leader Board to all players based on who have the
    most points required to achieve a goal.
    In this type of leader board, the lower the number the better.
    """
    score_type = EncodeScore(ScoreType)
    label = EncodeString(label)
    return Action(0, label, 0, 0, 0, goal, score_type, 37, 0, 4)


def MoveLocation(  # noqa: N802
    location: Location, on_unit: Unit, owner: Player, dest_location: Location
) -> Action:
    """Center location labeled location on units owned by player at location.

    This action will center a Location on a unit.
    In addition to choosing a location to move, you must specify a search
    location. The Action will ignore any units outside the search location.
    If no unit is found, the Location will move to the center of the search
    location. You can combine this Action with Center View to center the screen
    on a particular unit.
    """
    location = EncodeLocation(location)
    on_unit = EncodeUnit(on_unit)
    owner = EncodePlayer(owner)
    dest_loc = EncodeLocation(dest_location)
    return Action(dest_loc, 0, 0, 0, owner, location, on_unit, 38, 0, 20)


def MoveUnit(  # noqa: N802
    count: Count,
    unit_type: Unit,
    owner: Player,
    start_location: Location,
    dest_location: Location,
) -> Action:
    """Move quantity units for player at location to destination.

    This action will teleport a specified number of units (or unit) from one
    Location to another.
    """
    count = EncodeCount(count)
    unit = EncodeUnit(unit_type)
    owner = EncodePlayer(owner)
    start_loc = EncodeLocation(start_location)
    dest_loc = EncodeLocation(dest_location)
    return Action(start_loc, 0, 0, 0, owner, dest_loc, unit, 39, count, 20)


def LeaderBoardGreed(goal: Dword) -> Action:  # noqa: N802
    """Show Greed Leader Board for player closest to accumulation of number
    ore and gas.

    This will display the Leader Board to all players based on who is closest
    to reaching the goal of accumulating the most ore and gas.
    """
    return Action(0, 0, 0, 0, 0, goal, 0, 40, 0, 4)


def SetNextScenario(scenario_name: String) -> Action:  # noqa: N802
    """Load scenario after completion of current game.

    This trigger offers the ability to link multiple user-created maps together
    to form one large campaign.
    """
    scenario_name = EncodeString(scenario_name)
    return Action(0, scenario_name, 0, 0, 0, 0, 0, 41, 0, 4)


def SetDoodadState(  # noqa: N802
    state: PropState, unit: Unit, owner: Player, where: Location
) -> Action:
    """Set doodad state for units for player at location.

    The Installation tileset contains several doodads that can be enabled or
    disabled. The doors and concealed turrets can be set to start in one state
    or another by double clicking on them in the main window, but this action
    allows you to change their state during the course of the scenario. A
    location must be drawn around the doodads that you wish to affect with this
    action.

    Enabling a door closes it, and enabling a turret causes it to
    activate and attack any enemies of the trigger owner.
    """
    state = EncodePropState(state)
    unit = EncodeUnit(unit)
    owner = EncodePlayer(owner)
    where = EncodeLocation(where)
    return Action(where, 0, 0, 0, owner, 0, unit, 42, state, 20)


def SetInvincibility(  # noqa: N802
    state: PropState, unit: Unit, owner: Player, where: Location
) -> Action:
    """Set invincibility for units owned by player at location.

    This action makes the specified unit or units Invincible.
    Invincible units cannot be targeted or attacked, and take no damage.
    """
    state = EncodePropState(state)
    unit = EncodeUnit(unit)
    owner = EncodePlayer(owner)
    where = EncodeLocation(where)
    return Action(where, 0, 0, 0, owner, 0, unit, 43, state, 20)


def CreateUnit(  # noqa: N802
    number: Byte, unit: Unit, where: Location, for_player: Player
) -> Action:
    """Create quantity unit at location for player.

    This action creates the specified number of units at the specified
    Location. If a unit is created 'Anywhere', it will appear in the center of
    the map. This action will not function while the game is paused.

    Keep in mind that when the conditions are successfully met,
    the unit(s) will be created for each player that owns the trigger.
    For example, if **All Players** own a trigger that creates a Terran Marine
    for Player 1, and the conditions of the trigger are true for four of the
    players, Player 1 will get 4 Marines.
    """
    unit = EncodeUnit(unit)
    where = EncodeLocation(where)
    for_player = EncodePlayer(for_player)
    return Action(where, 0, 0, 0, for_player, 0, unit, 44, number, 20)


def SetDeaths(  # noqa: N802
    player: Player, modifier: Modifier, number: Dword, unit: Unit
) -> Action:
    """Modify death counts for player: Set quantity for unit.

    This will set the death counter of a particular unit, for the specified
    player, to a value listed in the action.
    """
    player = EncodePlayer(player)
    modifier = EncodeModifier(modifier)
    unit = EncodeUnit(unit)
    return Action(0, 0, 0, 0, player, number, unit, 45, modifier, 20)


def Order(  # noqa: N802
    unit: Unit,
    owner: Player,
    start_location: Location,
    order_type: _Order,
    dest_location: Location,
) -> Action:
    """Issue order to all units owned by player at location: order to
    destination.

    This action allows you to issue orders through a trigger to a unit
    (or units) that will change their behavior in a scenario.
    The different orders are attack, move and patrol.
    """
    unit = EncodeUnit(unit)
    owner = EncodePlayer(owner)
    start_loc = EncodeLocation(start_location)
    order_type = EncodeOrder(order_type)
    dest_loc = EncodeLocation(dest_location)
    return Action(start_loc, 0, 0, 0, owner, dest_loc, unit, 46, order_type, 20)


def Comment(text: String) -> Action:  # noqa: N802
    """Comment: comment.

    If this action exists in a trigger, and is enabled, whatever text is listed
    in the text field will be displayed in the trigger text.
    If you disable this action, the normal trigger text will be displayed.
    """
    text = EncodeString(text)
    return Action(0, text, 0, 0, 0, 0, 0, 47, 0, 4)


def GiveUnits(  # noqa: N802
    count: Count, unit: Unit, owner: Player, where: Location, new_owner: Player
) -> Action:
    """Give quantity units owned by player at location to player.

    This action allows you to transfer units from one player to another.
    """
    count = EncodeCount(count)
    unit = EncodeUnit(unit)
    owner = EncodePlayer(owner)
    where = EncodeLocation(where)
    new_owner = EncodePlayer(new_owner)
    return Action(where, 0, 0, 0, owner, new_owner, unit, 48, count, 20)


def ModifyUnitHitPoints(  # noqa: N802
    count: Count, unit: Unit, owner: Player, where: Location, percent: Dword
) -> Action:
    """Set hit points for quantity units owned by player at location to
    percent%.

    This action will modify the specified unit(s) hit points. The hit points
    will be changed based on the percentage specified in the action trigger.
    """
    count = EncodeCount(count)
    unit = EncodeUnit(unit)
    owner = EncodePlayer(owner)
    where = EncodeLocation(where)
    return Action(where, 0, 0, 0, owner, percent, unit, 49, count, 20)


def ModifyUnitEnergy(  # noqa: N802
    count: Count, unit: Unit, owner: Player, where: Location, percent: Dword
) -> Action:
    """Set energy points for quantity units owned by player at location to
    percent%.

    This action will modify the specified unit(s) spell-casting energy. The
    energy will be changed based on the percentage specified in the action
    trigger.
    """
    count = EncodeCount(count)
    unit = EncodeUnit(unit)
    owner = EncodePlayer(owner)
    where = EncodeLocation(where)
    return Action(where, 0, 0, 0, owner, percent, unit, 50, count, 20)


def ModifyUnitShields(  # noqa: N802
    count: Count, unit: Unit, owner: Player, where: Location, percent: Dword
) -> Action:
    """Set shield points for quantity units owned by player at location to
    percent%.

    This action will modify the specified unit(s) shield points.
    The shield points will be changed based on the percentage specified in the
    action trigger.
    """
    count = EncodeCount(count)
    unit = EncodeUnit(unit)
    owner = EncodePlayer(owner)
    where = EncodeLocation(where)
    return Action(where, 0, 0, 0, owner, percent, unit, 51, count, 20)


def ModifyUnitResourceAmount(  # noqa: N802
    count: Count, owner: Player, where: Location, new_value: Dword
) -> Action:
    """Set resource amount for quantity resource sources owned by player at
    location to quantity.

    This action allows you to modify the amount of resources contained in the
    various mineral stores. For example, you could modify a Vespene Geyser
    so that it had 0 resources if you desire.
    """
    count = EncodeCount(count)
    owner = EncodePlayer(owner)
    where = EncodeLocation(where)
    return Action(where, 0, 0, 0, owner, new_value, 0, 52, count, 4)


def ModifyUnitHangarCount(  # noqa: N802
    add: Dword, count: Count, unit: Unit, owner: Player, where: Location
) -> Action:
    """Add at most quantity to hangar for quantity units at location owned by
    player.

    This action will modify the contents of a unit(s) hangar.
    For example, this will allow you to add 5 additional Interceptors to the
    Carrier's hangar.
    """
    count = EncodeCount(count)
    unit = EncodeUnit(unit)
    owner = EncodePlayer(owner)
    where = EncodeLocation(where)
    return Action(where, 0, 0, 0, owner, add, unit, 53, count, 20)


def PauseTimer() -> Action:  # noqa: N802
    """Pause the game.

    This action will put the game in a pause state.
    If a matching Unpause Game is not found, the program automatically unpauses
    the game when the current trigger is finished. Note that pause game has
    no effect in multiplayer scenarios or against computer controlled players.
    """
    return Action(0, 0, 0, 0, 0, 0, 0, 54, 0, 4)


def UnpauseTimer() -> Action:  # noqa: N802
    """Unpause the countdown timer.

    This action will resume the timer from a paused session.
    """
    return Action(0, 0, 0, 0, 0, 0, 0, 55, 0, 4)


def Draw() -> Action:  # noqa: N802
    """End the scenario in a draw for all players.

    This will end the scenario in a draw for the affected players.
    Any other players in the game will continue.
    """
    return Action(0, 0, 0, 0, 0, 0, 0, 56, 0, 4)


def SetAllianceStatus(player: Player, status: AllyStatus) -> Action:  # noqa: N802
    """Set Player to Ally status.

    This allows you to set the value of the affected players' alliance status.
    """
    player = EncodePlayer(player)
    ally = EncodeAllyStatus(status)
    return Action(0, 0, 0, 0, player, 0, ally, 57, 0, 4)


def SetMemory(dest: Dword, modtype: Modifier, value: Dword) -> Action:  # noqa: N802
    modifier = EncodeModifier(modtype)
    return Action(0, 0, 0, 0, EPD(dest), value, 0, 45, modifier, 20)


def SetMemoryEPD(dest: Dword, modtype: Modifier, value: Dword) -> Action:  # noqa: N802
    epd = EncodePlayer(dest)
    modifier = EncodeModifier(modtype)
    return Action(0, 0, 0, 0, epd, value, 0, 45, modifier, 20)


def SetNextPtr(trg: Dword, dest: Dword) -> Action:  # noqa: N802
    return SetMemory(trg + 4, 7, dest)


def SetDeathsX(  # noqa: N802
    player: Player, modifier: Modifier, number: Dword, unit: Unit, mask: Dword
) -> Action:
    player = EncodePlayer(player)
    modifier = EncodeModifier(modifier)
    unit = EncodeUnit(unit)
    return Action(mask, 0, 0, 0, player, number, unit, 45, modifier, 20, eudx=0x4353)


def SetMemoryX(  # noqa: N802
    dest: Dword, modtype: Modifier, value: Dword, mask: Dword
) -> Action:
    modifier = EncodeModifier(modtype)
    return SetDeathsX(EPD(dest), modifier, value, 0, mask)


def SetMemoryXEPD(  # noqa: N802
    epd: Dword, modtype: Modifier, value: Dword, mask: Dword
) -> Action:
    modifier = EncodeModifier(modtype)
    return SetDeathsX(epd, modifier, value, 0, mask)


def SetKills(  # noqa: N802
    player: Player, modifier: Modifier, number: Dword, unit: Unit
) -> Action | tuple[Action, Action, Action]:
    player = EncodePlayer(player)
    modifier = EncodeModifier(modifier)
    unit = EncodeUnit(unit)
    if isinstance(player, int) and player >= 12:
        if player == 13:
            return (
                SetMemory(0x6509B0, Add, -12 * 228),
                SetDeaths(13, modifier, number, unit),  # CurrentPlayer
                SetMemory(0x6509B0, Add, 12 * 228),
            )
        ep_assert(
            player <= 11,
            _("SetKills Player should be only P1~P12 or CurrentPlayer"),
        )
    if isinstance(unit, int):
        ep_assert(unit <= 227, _("SetKills Unit should be at most 227"))
    return SetDeaths(player - 228 * 12, modifier, number, unit)



# Copyright 2014 by trgk.
# All rights reserved.
# This file is part of EUD python library (eudplib),
# and is released under "MIT License Agreement". Please see the LICENSE
# file that should have been included as part of this package.

from ctypes import LittleEndianStructure, c_byte, c_uint, c_ushort
from struct import pack, unpack_from

from ... import utils as ut


class UnitProperty(LittleEndianStructure):

    """
    UnitProperty class. Used in 'Create Unit with Properties' action.
    """

    _fields_ = [  # noqa: RUF012
        ("sprpvalid", c_ushort),
        ("prpvalid", c_ushort),
        ("player", c_byte),
        ("hitpoint", c_byte),
        ("shield", c_byte),
        ("energy", c_byte),
        ("resource", c_uint),
        ("hanger", c_ushort),
        ("sprpflag", c_ushort),
        ("unused", c_uint),
    ]

    def __init__(
        self,
        hitpoint: int | None = None,
        shield: int | None = None,
        energy: int | None = None,
        resource: int | None = None,
        hanger: int | None = None,
        cloaked: bool | None = None,
        burrowed: bool | None = None,
        intransit: bool | None = None,
        hallucinated: bool | None = None,
        invincible: bool | None = None,
    ) -> None:
        """
        Properties : Value/None (Don't care)

        - hitpoint : 0~100(%)  if) When unit's hitpoint is greater than 167772,
        - shield   : 0~100(%)   you should give hitpoint None to make 100%% HP.
        - energy   : 0~100(%)
        - resource : 0~4294967295
        - hanger   : 0~65536 (count)

        Special properties : True(Enabled)/False(Disabled)/None(Don't care)

        - clocked      : Unit is clocked.
        - burrowed     : Unit is burrowed.
        - intransit    : Unit is lifted. (In transit)
        - hallucinated : Unit is hallucination.
        - invincible   : Unit is invincible.

        >>> UnitProperty(hitpoint = 50, burrowed = True) # HP 50%, burrowed
        """
        ut.ep_assert(
            hitpoint is None or (isinstance(hitpoint, int) and 0 <= hitpoint <= 100)
        )
        ut.ep_assert(
            shield is None or (isinstance(shield, int) and 0 <= shield <= 100)
        )
        ut.ep_assert(
            energy is None or (isinstance(energy, int) and 0 <= energy <= 100)
        )
        ut.ep_assert(
            resource is None
            or (isinstance(resource, int) and 0 <= resource <= 65535)
        )
        ut.ep_assert(
            hanger is None or (isinstance(hanger, int) and 0 <= hanger <= 255)
        )

        ut.ep_assert(cloaked in [None, True, False])
        ut.ep_assert(burrowed in [None, True, False])
        ut.ep_assert(intransit in [None, True, False])
        ut.ep_assert(hallucinated in [None, True, False])
        ut.ep_assert(invincible in [None, True, False])

        def prop2int(p):
            if p is None:
                return 0
            else:
                return p

        def prop2valid(p, v):
            if p is None:
                return 0
            else:
                return v

        def prop2flag(p, v):
            if p:
                return v
            else:
                return 0

        self.player = 0

        # Set properties
        self.hitpoint = prop2int(hitpoint)
        self.shield = prop2int(shield)
        self.energy = prop2int(energy)
        self.resource = prop2int(resource)
        self.hanger = prop2int(hanger)

        self.prpvalid = (
            prop2valid(hitpoint, 1 << 1)
            | prop2valid(shield, 1 << 2)
            | prop2valid(energy, 1 << 3)
            | prop2valid(resource, 1 << 4)
            | prop2valid(hanger, 1 << 5)
        )

        # Set special properties
        self.sprpvalid = (
            prop2valid(cloaked, 1 << 0)
            | prop2valid(burrowed, 1 << 1)
            | prop2valid(intransit, 1 << 2)
            | prop2valid(hallucinated, 1 << 3)
            | prop2valid(invincible, 1 << 4)
        )

        self.sprpflag = (
            prop2flag(cloaked, 1 << 0)
            | prop2flag(burrowed, 1 << 1)
            | prop2flag(intransit, 1 << 2)
            | prop2flag(hallucinated, 1 << 3)
            | prop2flag(invincible, 1 << 4)
        )


def property_key(b: bytes, index: int = 0) -> bytes:
    x = list(unpack_from("<HHBBBBIHHI", b, index))

    x[0] &= 0b11111  # remove unused flags
    x[1] &= 0b111111  # remove unused flags
    x[8] &= 0b11111  # remove unused flags

    x[2] = 0  # player number, always NULL
    x[9] = 0  # padding?

    if x[3] == 100:  # 100% HP is equals to not using HP%
        x[1] &= ~0b10
    if x[4] == 100:  # 100% Shields is equals to not using Shields%
        x[1] &= ~0b100
    # Can't apply to Energy: Default energy differs between max energy (200 and 250)
    # Can't apply to Resource amount: Default amount differs
    # between Mineral Fields and Vespene Geyser
    if x[7] == 0:  # 0 in hanger is equals to not using hanger amount
        x[1] &= ~0b100000

    if x[1] & 0b10 == 0:  # HP is not valid
        x[3] = 0
    if x[1] & 0b100 == 0:  # Shields is not valid
        x[4] = 0
    if x[1] & 0b1000 == 0:  # Energy is not valid
        x[5] = 0
    if x[1] & 0b10000 == 0:  # Resource amount is not valid
        x[6] = 0
    if x[1] & 0b100000 == 0:  # Amount in hanger is not valid
        x[7] = 0

    if x[8] & 0b1 == 0:  # No Cloak is equals to not using Cloak
        x[0] &= ~0b1
    if x[8] & 0b10 == 0:  # No Burrowed is equals to not using Burrowed
        x[0] &= ~0b10
    if x[8] & 0b100 == 0:  # No Lifted is equals to not using Lifted
        x[0] &= ~0b100
    if x[8] & 0b1000 == 0:  # No Hallucinated is equals to not using Hallucinated
        x[0] &= ~0b1000
    if x[8] & 0b10000 == 0:  # No Invincible is equals to not using Invincible
        x[0] &= ~0b10000

    if x[0] & 0b1 == 0:  # Cloak is not valid
        x[8] &= ~0b1
    if x[0] & 0b10 == 0:  # Burrowed is not valid
        x[8] &= ~0b10
    if x[0] & 0b100 == 0:  # In transit is not valid
        x[8] &= ~0b100
    if x[0] & 0b1000 == 0:  # Hallucinated is not valid
        x[8] &= ~0b1000
    if x[0] & 0b10000 == 0:  # Invincible is not valid
        x[8] &= ~0b10000

    return pack("<HHBBBBIHHI", *x)



from collections import deque

from helper import *

from .test_eps_array import (
    DequeCases,
    DequeTest,
    f_test_alias,
    f_test_compare,
    f_test_deque,
    f_test_queue,
    f_test_queue_wraparound,
    f_test_write,
)
from .test_eps_compat import f_test_compatibility, stat
from .test_eps_cunit import f_test_cunit1, f_test_cunit2
from .test_eps_epsfile import (
    f_constv_thing,
    f_square,
    f_test_reported,
    test_eps_misc,
)
from .test_eps_object import (
    f_test_eudmethods,
    f_test_nested_object,
    f_test_object,
    f_test_selftype_member,
)


@TestInstance
def test_epscript():
    test_equality("epScript compile 1", f_square(4), 16)
    test_equality("epScript compile 2", f_constv_thing(), 55)
    a = test_eps_misc.f_switch_test()
    test_equality(
        "epScript switch",
        [a[0], a[1], a[2], a[3], a[4], a[5]],
        [1239, 1243, 1246, 1256, 1258, 1260],
    )
    test_equality("epScript array", test_eps_misc.f_test_array(), 23)
    v = f_test_write()
    test_equality(
        "epScript array write",
        v[1:],
        [v[0]] * (len(v) - 1),
    )
    test_eps_misc.f_test_all_actions()
    c = f_test_compare()
    test_equality(
        "epScript array compare",
        c[1:],
        [c[0]] * (len(c) - 1),
    )
    test_equality("epScript EUDQueue", f_test_queue(), 31)
    test_equality(
        "epScript queue wraparound",
        f_test_queue_wraparound(),
        [0, 3, 9, 12, 9, 20, 14],
    )
    dq = deque(maxlen=7)
    cases = []
    pushes = 1
    for method in DequeCases:
        if method in ("append", "appendleft"):
            f = getattr(dq, method)
            f(pushes)
            pushes += 1
        elif dq:
            f = getattr(dq, method)
            cases.append(f())
        else:
            cases.append(0)
    cases.append(len(dq) + sum(dq))
    testname = DequeTest + "".join(
        chr(14 + i % 4) + str(x) for i, x in enumerate(cases)
    )
    test_equality("epScript EUDDeque %s" % testname, f_test_deque(), cases)
    arr = f_test_alias()
    test_equality(
        "epScript array alias write a[i] = i",
        [arr[i] for i in range(10)],
        list(range(10)),
    )

    test_equality("epScript object", f_test_object(), 511)
    test_equality("epScript nested object", f_test_nested_object(), 127)
    test_equality("epScript all object array", f_test_selftype_member(), 5)
    test_equality("epScript EUDMethod", f_test_eudmethods(), 7)
    test_equality("epScript EPDCUnitMap", f_test_cunit1(), 0x7FFFF)
    test_equality("epScript CUnit", f_test_cunit2(), 8191)
    test_equality("epScript compatibility", f_test_compatibility(), 32)
    test_equality(
        "epScript stat_txt.tbl",
        stat.f_test_stattext(),
        [ord(c) for c in stat.expected_result],
    )
    test_equality("test updateUnitNameAndRank", f_test_reported(), [0, 0])
